from cloudsend import provider as cs
from cloudsend.provider import COMMAND_ARGS_SEP , ARGS_SEP
import asyncio , os , sys , time
from cloudsend.core import CloudSendProcessState , bcolors
import traceback
import json
import socket
import asyncio

HOST = 'localhost' #'0.0.0.0' #127.0.0.1' 
PORT = 5000


async def mainloop(ctxt):

    await ctxt.init()

    server = await asyncio.start_server(ctxt.handle_client, HOST, PORT, reuse_address=True, reuse_port=True)

    addrs = ', '.join(str(sock.getsockname()) for sock in server.sockets)
    print(f'Serving on {addrs}')

    try:
        async with server:
            await server.serve_forever()   
    except asyncio.CancelledError:
        ctxt.restore_stdio()
        print("Shutting down ...")


class ByteStreamWriter():

    def __init__(self,writer):
        self.writer = writer 
    
    def write(self,data):
        if isinstance(data,(bytes,bytearray)):
            self.writer.write(data)
        elif isinstance(data,str):
            self.writer.write(data.encode('utf-8'))
        else:
            self.writer.write(str(data,'utf-8').encode('utf-8'))

    async def drain(self):
        await self.writer.drain()

    def close(self):
        self.writer.close()

    def flush(self):
        asyncio.ensure_future( self.writer.drain() )

class ServerContext:
    def __init__(self,args):
        self.cs_client  = None
        self.old_stdout = sys.stdout
        self.old_stderr = sys.stderr
        self.auto_init = any( arg == 'auto_init' for arg in args )

    async def init(self):
        if self.auto_init: # set to True by the crontab in case maestro crashed
            self.cs_client = cs.get_client(None)
            # we've just restarted a crashed maestro server process
            # let's test for WATCH state and reach it back again if thats needed
            await self.wakeup()

    async def wakeup(self):
        await self.cs_client.wakeup()

    async def handle_client(self, reader, writer):
        while True:
            try:
                cmd_line = await reader.readline()
                if not cmd_line:
                    break
                cmd_line = cmd_line.decode('utf-8').strip()
                cmd_args = cmd_line.split(COMMAND_ARGS_SEP)
                if len(cmd_args)>=2:
                    cmd  = cmd_args[0].strip()
                    args = cmd_args[1].split(ARGS_SEP)
                else:
                    cmd  = cmd_line
                    args = None

                await self.process_command(cmd,args,writer)
                break # one-shot command
            except ConnectionResetError as cre:
                try:
                    sys.stdout = old_stdout
                    print(cre)
                    print("DISCONNECTION")
                except:
                    pass
                break    
            except Exception as e:
                try:
                    sys.stdout = old_stdout
                    print(e)
                except:
                    pass
                break  
        try:
            await writer.drain()
        except:
            pass
        try:
            writer.close()
        except:
            pass

    async def process_command(self,command,args,writer):
        #sock = writer.transport.get_extra_info('socket')
        string_writer = ByteStreamWriter(writer)
        sys.stdout = string_writer
        sys.stderr = string_writer

        if self.cs_client is None:
            if command != 'init' and command != 'shutdown':
                print(bcolors.WARNING+"Server not ready. Run 'init CONFIG_FILE' or 'start' command first"+bcolors.ENDC)
                await writer.drain()
                return 
        try:

            if command == 'init':

                if not args:
                    config_ = None
                elif len(args)==1:
                    config_ = args[0].strip()
                else:
                    config_ = None
                self.cs_client  = cs.get_client(config_)

                await self.wakeup()                

            elif command == 'wakeup':

                await self.cs_client.wakeup()

            elif command == 'start':

                if not args:
                    reset = False
                elif len(args)==1:
                    reset = args[0].strip().lower() == "true"
                else:
                    reset = False

                await self.cs_client.start(reset)

            elif command == 'cfg_add_instances':
                if len(args)==1:
                    config = args[0]
                else:
                    print("Error: you need to send a JSON stream for config")
                    await writer.drain()
                    return
                await self.cs_client.cfg_add_instances(config)

            elif command == 'cfg_add_environments':
                if len(args)==1:
                    config = args[0]
                else:
                    print("Error: you need to send a JSON stream for config")
                    await writer.drain()
                    return
                await self.cs_client.cfg_add_environments(config)

            elif command == 'cfg_add_jobs':
                if len(args)==1:
                    config = args[0]
                else:
                    print("Error: you need to send a JSON stream for config")
                    await writer.drain()
                    return
                await self.cs_client.cfg_add_jobs(config)

            elif command == 'cfg_add_config':
                if len(args)==1:
                    config = json.loads( args[0] )
                else:
                    print("Error: you need to send a JSON stream for config")
                    await writer.drain()
                    return
                await self.cs_client.cfg_add_config(config)

            elif command == 'cfg_reset':

                await self.cs_client.cfg_reset()

            elif command == 'deploy':

                await self.cs_client.deploy()

            elif command == 'run':

                await self.cs_client.run()
            
            elif command == 'wait':

                await self.cs_client.wait(CloudSendProcessState.DONE|CloudSendProcessState.ABORTED)

            elif command == 'get_states':

                await self.cs_client.get_jobs_states()

            elif command == 'print_summary' or command == 'print':

                await self.cs_client.print_jobs_summary()

            elif command == 'print_aborted':

                await self.cs_client.print_aborted_logs()

            elif command == 'print_objects':

                await self.cs_client.print_objects()

            elif command == 'fetch_results':

                if args:
                    directory = args[0].strip()
                    await self.cs_client.fetch_results(directory)

            elif command == 'shutdown':

                asyncio.get_event_loop().stop()

            elif command == 'test':

                print("TEST")
            
            else:

                print("UNKNOWN COMMAND")
        
        except Exception as e:
            print(e)
            await writer.drain()
            #writer.close()
            self.restore_stdio()
            print(e)
            raise e

        await writer.drain()
        # to let time for the buffer to be flushed before killing thread and connection
        #time.sleep(1)
        #writer.close()  

    def restore_stdio(self):
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr

# run main loop
def main():

    ctxt = ServerContext(sys.argv)

    try:
        asyncio.run( mainloop(ctxt) )
    except RuntimeError:
        pass

if __name__ == '__main__':
    main()    


# combinatorial optimization

# bin packing
from ortools.linear_solver import pywraplp
from ortools.sat.python import cp_model

# https://developers.google.com/optimization/bin/multiple_knapsack

def multiple_knapsack_assignation(jobs,instances,method='mip'):

    assigned = [False]*len(jobs)

    for i,job in enumerate(jobs):
        if job.get_instance() is not None:
            assigned[i] = True

    while True:

        if method == 'mip':
            res , indexmap = _multiple_knapsack_assignation_pass_MIP(jobs,instances,assigned)

        elif method == 'cp_sat':
            res , indexmap = _multiple_knapsack_assignation_pass_CP_SAT(jobs,instances,assigned)

        for b in range(len(instances)):
            for i in range(len(indexmap)):
                if res[i,b]:
                    assigned[indexmap[i]] = True
                    print("job",indexmap[i],">>>> instance",b)
                    jobs[indexmap[i]].set_instance(instances[b])

        num_assigned = sum(assigned)

        if num_assigned == len(jobs):
            break


def _multiple_knapsack_fill_up_data(jobs,instances,assigned):
    data = {}
    data['weights'] = []
    data['values']  = []
    indexmap = []
    for i,job in enumerate(jobs):
        # the job has already been assigned in a previous pass
        if assigned[i]:
            continue

        cpu_req = job.get_config('cpu_reqs')
        if cpu_req is None:
            cpu_req = 1
        weight = cpu_req
        value  = cpu_req
        # record the mapping
        indexmap.append(i)
        data['weights'].append(weight)
        data['values'].append(value)

    assert len(data['weights']) == len(data['values'])
    data['num_items'] = len(data['weights'])
    data['all_items'] = range(data['num_items'])

    data['bin_capacities'] = []
    for instance in instances:
        cpus = instance.get_config('cpus')
        if cpus is None:
            cpus = 1 #TODO: use the spec from the type ....
        data['bin_capacities'].append(cpus)

    data['num_bins'] = len(data['bin_capacities'])
    data['all_bins'] = range(data['num_bins'])
    return data , indexmap

# THIS ONE is Seg Faulting ???

def _multiple_knapsack_assignation_pass_MIP(jobs,instances,assigned):

    # FILL UP the data

    data , indexmap = _multiple_knapsack_fill_up_data(jobs,instances,assigned)

   
    # DECLARE the Solver

    solver = pywraplp.Solver.CreateSolver('SCIP')
    if solver is None:
        print('SCIP solver unavailable.')
        return 

    
    # CREATE the variables

    # x[i, b] = 1 if item i is packed in bin b.
    x = {}
    for i in data['all_items']:
        for b in data['all_bins']:
            x[i, b] = solver.BoolVar(f'x_{i}_{b}')

    
    # DEFINE the constraints

    # Each item is assigned to at most one bin.
    for i in data['all_items']:
        solver.Add(sum(x[i, b] for b in data['all_bins']) <= 1)

    # The amount packed in each bin cannot exceed its capacity.
    for b in data['all_bins']:
        solver.Add(
            sum(x[i, b] * data['weights'][i]
                for i in data['all_items']) <= data['bin_capacities'][b])    


    # DEFINE the objective

    # Maximize total value of packed items.
    objective = solver.Objective()
    for i in data['all_items']:
        for b in data['all_bins']:
            objective.SetCoefficient(x[i, b], data['values'][i])
    objective.SetMaximization()    

    status = solver.Solve()

    # PRINT the solution 

    result = {}

    if status == pywraplp.Solver.OPTIMAL:
        #print(f'Total packed value: {objective.Value()}')
        total_weight = 0
        for b in data['all_bins']:
            #print(f'Bin {b}')
            bin_weight = 0
            bin_value = 0
            for i in data['all_items']:
                if x[i, b].solution_value() > 0:
                    #print(
                    #    f"Item {i} weight: {data['weights'][i]} value: {data['values'][i]}"
                    #)
                    bin_weight += data['weights'][i]
                    bin_value += data['values'][i]
                    result[i,b] = True
                else:
                    result[i,b] = False
            #print(f'Packed bin weight: {bin_weight}')
            #print(f'Packed bin value: {bin_value}\n')
            total_weight += bin_weight
        #print(f'Total packed weight: {total_weight}')
    else:
        print('The problem does not have an optimal solution.')    

    return result , indexmap 

def _multiple_knapsack_assignation_pass_CP_SAT(jobs,instances,assigned):

    # DECLARE the model

    model = cp_model.CpModel()

    # CREATE the data

    data , indexmap = _multiple_knapsack_fill_up_data(jobs,instances,assigned)

    # CREATE the variables

    # x[i, b] = 1 if item i is packed in bin b.
    x = {}
    for i in data['all_items']:
        for b in data['all_bins']:
            x[i, b] = model.NewBoolVar(f'x_{i}_{b}')

    # ADD constraints 
    # Each item is assigned to at most one bin.
    for i in data['all_items']:
        model.AddAtMostOne(x[i, b] for b in data['all_bins'])

    # The amount packed in each bin cannot exceed its capacity.
    for b in data['all_bins']:
        model.Add(
            sum(x[i, b] * data['weights'][i]
                for i in data['all_items']) <= data['bin_capacities'][b])

    # Objective.
    # Maximize total value of packed items.
    objective = []
    for i in data['all_items']:
        for b in data['all_bins']:
            objective.append(
                cp_model.LinearExpr.Term(x[i, b], data['values'][i]))
    model.Maximize(cp_model.LinearExpr.Sum(objective))

    solver = cp_model.CpSolver()
    status = solver.Solve(model)

    result = {}

    if status == cp_model.OPTIMAL:
        #print(f'Total packed value: {solver.ObjectiveValue()}')
        total_weight = 0
        for b in data['all_bins']:
            #print(f'Bin {b}')
            bin_weight = 0
            bin_value = 0
            for i in data['all_items']:
                if solver.Value(x[i, b]) > 0:
                    #print(
                    #    f"Item {i} weight: {data['weights'][i]} value: {data['values'][i]}"
                    #)
                    bin_weight += data['weights'][i]
                    bin_value += data['values'][i]
                    result[i,b] = True
                else:
                    result[i,b] = False
            #print(f'Packed bin weight: {bin_weight}')
            #print(f'Packed bin value: {bin_value}\n')
            total_weight += bin_weight
        #print(f'Total packed weight: {total_weight}')
    else:
        print('The problem does not have an optimal solution.')

    return result , indexmap



from cloudsend import provider as cs
import asyncio , os , sys
from cloudsend.core import CloudSendProcessState
import traceback
import json

async def tail_loop(script_hash,uid):

    generator = await cs_client.tail(script_hash,uid) 
    for line in generator:
        print(line)


async def mainloop(cs_client,reset=False):

    print("\n== START ==\n")

    # distribute the jobs on the instances (dummy algo for now)
    await cs_client.start(reset) 

    print("\n== DEPLOY ==\n")

    # pre-deploy instance , environments and job files
    # it is recommended to wait here allthough run.sh should wait for bootstraping
    # currently, the bootstraping is non-blocking
    # so this will barely wait ... (the jobs will do the waiting ...)
    await cs_client.deploy()

    print("\n== RUN ==\n")

    # run the scripts and get a process back
    await cs_client.run()

    print("\n== WAIT ==\n")

    print("Waiting for DONE or ABORTED ...")
    # now that we have 'watch' before 'wait' , this will exit instantaneously
    # because watch includes 'wait' mode intrinsiquely
    await cs_client.wait(CloudSendProcessState.DONE|CloudSendProcessState.ABORTED)

    print("\n== SUMMARY ==\n")

    # just to show the API ...
    await cs_client.get_jobs_states()

    # print("\n== WAIT and TAIL ==\n")

    await cs_client.print_aborted_logs()

    print("\n== FETCH RESULTS ==\n")

    await cs_client.fetch_results(os.path.join(os.getcwd(),'tmp'))

    print("\n== DONE ==\n")

async def waitloop(cs_client):

    print("\n== START ==\n")

    await cs_client.start()

    print("\n== WAIT ==\n")
    
    await cs_client.wait(CloudSendProcessState.DONE|CloudSendProcessState.ABORTED)

    print("\n== SUMMARY ==\n")

    # just to show the API ...
    await cs_client.get_jobs_states()
    await cs_client.print_aborted_logs()

    print("\n== FETCH RESULTS ==\n")

    await cs_client.fetch_results(os.path.join(os.getcwd(),'tmp'))

    print("\n== DONE ==\n")        

# run main loop
def main():

    if len(sys.argv)<=1:
        print("You need to specify a config file path: config.json or config.py")
        sys.exit(1)
    
    config_file = sys.argv[1]
    cs_client   = cs.get_client(config_file)
    command = None
    if len(sys.argv)>2: 
        command = sys.argv[2]
    
    if command == 'wait':
        asyncio.run( waitloop(cs_client) )
    elif command == 'reset':
        asyncio.run( mainloop(cs_client,True) )
    else:
        asyncio.run( mainloop(cs_client) )

if __name__ == '__main__':
    main()    


import hashlib , os , subprocess , json
import jcs , yaml
import uuid
import os
from cloudsend.core import K_COMPUTED

# keys used for hash computation
# Note: we include market options (SPOT ON|OFF e.g.) for the instance because it defines how the 'hardware' will run 
#       so it's considered part of the intrinsic characteristics of the machine
cs_instance_keys       = [ 'region'  , 'cloud_id' , 'img_id' , 'type' , 'cpus' , 'gpu' , 'disk_size' , 'disk_type' , 'eco' , 'max_bid' ] 
cs_environment_keys    = [ 'command' , 'env_pypi' , 'env_conda' , 'env_aptget' , 'env_julia' ]

def compute_instance_hash(instance_cfg):
    instance_config = { your_key: instance_cfg.get(your_key) for your_key in cs_instance_keys }
    instance_config_canon = jcs.canonicalize(instance_config)
    #print("INSTANCE CONFIG CANON",instance_config_canon)
    #instance_hash = str(hash(instance_config_canon))
    hash = hashlib.md5(instance_config_canon).hexdigest()
    return hash[0:12]


def add_pip_dependency_to_conda(env_dict):
    if not 'dependencies' in env_dict['env_conda']:
        env_dict['env_conda']['dependencies'] = [] 
    added = False
    pipIndex = -1 
    for i,dep in enumerate(env_dict['env_conda']['dependencies']):
        if dep == 'pip':
            env_dict['env_conda']['dependencies'][i] = { 'pip' : ['-r __REQUIREMENTS_TXT_LINK__'] }
            added = True
            pipIndex = i 
            break
        elif isinstance(dep,dict) and 'pip' in dep and isinstance(dep['pip'],list):
            pipIndex = i 
            doBreak = False
            for pipdep in dep['pip']:
                if 'requirements.txt' in pipdep:
                    doBreak = True
                    added   = True 
                    break
            if doBreak:
                break 
    # add the requirements dependecy (it was not found)
    if added is False:
        if pipIndex != -1:
            env_dict['env_conda']['dependencies'][pipIndex]['pip'].append('-r __REQUIREMENTS_TXT_LINK__')
        else:
            env_dict['env_conda']['dependencies'].append({'pip':['-r __REQUIREMENTS_TXT_LINK__']})

def update_requirements_path(env_dict,path):
    if isinstance(env_dict,dict):
        for k,v in env_dict.items():
            env_dict[k] = update_requirements_path(v,path)
        

    elif isinstance(env_dict,list):
        for i,v in enumerate(env_dict):
            env_dict[i] = update_requirements_path(v,path)
    
    elif isinstance(env_dict,str):
        if '__REQUIREMENTS_TXT_LINK__' in env_dict:
            env_dict = env_dict.replace('__REQUIREMENTS_TXT_LINK__',path)
        
    return env_dict

# returns a JSON object that represents lists as in requirements.txt as well as YAML format (can be parsed by YAML module)
# this object will be serialized and sent to remote host
# this is also used to compute the hash for the environment
def compute_environment_object(env_config):

    # the configuration is already computed (usually by the LightProvider)
    # we just have to return this env object for deployment
    if env_config.get(K_COMPUTED) == True:
        return env_config

    environment_obj = {
        'command'    : None ,
        'env_aptget' : None , 
        'env_conda'  : None , 
        'env_pypi'   : None , 
        'env_julia'  : None ,
        'name'       : env_config.get('name')
    }
    
    if env_config.get('command'):

        env_command = env_config['command']

        if isinstance(env_command,str) and os.path.isfile(env_command):

            with open(env_command, "r") as bash_file:
                try:
                    environment_obj['command'] = bash_file.read()
                except Exception as exc:
                    print(exc)
            
    # conda+pip: https://stackoverflow.com/questions/35245401/combining-conda-environment-yml-with-pip-requirements-txt
    
    if env_config.get('env_aptget'):

        env_apt_get = env_config['env_aptget']

        if isinstance(env_apt_get,list):

            # set dependencies as this
            env_apt_get = list(map(str.strip,env_apt_get)) # strip the strings
            environment_obj['env_aptget'] = env_apt_get


    if env_config.get('env_conda'):

        env_conda = env_config['env_conda']

        if isinstance(env_conda,list):
        
            # sort the array and set dependencies as this
            env_conda = list(map(str.strip,env_conda)) # strip the strings
            environment_obj['env_conda'] = dict()
            environment_obj['env_conda']['dependencies'] = env_conda
            conda_channels = env_config.get('env_conda_channels')
            if not conda_channels:
                conda_channels = [ 'conda-forge' , 'defaults' ]
            environment_obj['env_conda']['channels'] = conda_channels
        
        elif isinstance(env_conda,str) and os.path.isfile(env_conda):

            with open(env_conda, "r") as ymlData:
                try:
                    ymlConda = yaml.load(ymlData,Loader=yaml.FullLoader)
                    environment_obj['env_conda'] = ymlConda
                except yaml.YAMLError as exc:
                    print(exc)
        
        elif isinstance(env_conda,str) and os.path.isdir(env_conda):

            cmd    = ['source activate "'+env_conda+'" && conda env export']
            try:
                result = subprocess.run(cmd, stdout=subprocess.PIPE,shell=True)
                try:
                    ymlConda = yaml.load(result.stdout,Loader=yaml.FullLoader)
                    environment_obj['env_conda'] = ymlConda
                except yaml.YAMLError as exc:
                    print(exc)
            except subprocess.CalledProcessError as pex:
                print(pex)
        
        else:
            print("\033[93menv_conda is specified but the file or directory doesn't exist\033[0m")


    if env_config.get('env_pypi'):

        env_pypi = env_config['env_pypi']

        #make sure we have the extra ref in the conda env, if conda env exists ...
        if environment_obj['env_conda'] is not None:
            add_pip_dependency_to_conda( environment_obj )

        if isinstance(env_pypi,list):

            # set dependencies as this
            env_pypi = list(map(str.strip,env_pypi)) # strip the strings
            environment_obj['env_pypi'] = env_pypi

        elif isinstance(env_pypi,str) and os.path.isfile(env_pypi):

            with open(env_pypi, "r") as txtFile:
                pipDeps = txtFile.read().split('\n')
                environment_obj['env_pypi'] = pipDeps
            
        elif isinstance(env_pypi,str) and os.path.isdir(env_pypi):

            #cmd    = "source " + os.path.realpath(env_pypi) + os.path.sep + 'bin' + os.path.sep + 'activate && pip freeze'
            cmd = [
                os.path.realpath(env_pypi) + os.path.sep + 'bin' + os.path.sep + 'pip' ,
                'freeze'
            ]
            try:
                result = subprocess.run(cmd, stdout=subprocess.PIPE)
                pipDeps = result.stdout.decode("utf-8").strip().split('\n')
                environment_obj['env_pypi'] = pipDeps
            except subprocess.CalledProcessError as pex:
                print(pex)

        else:
            print("\033[93menv_pypi is specified but the file or directory doesn't exist\033[0m")

    if env_config.get('env_julia'):

        env_julia = env_config['env_julia']

        if isinstance(env_julia,str):

            env_julia = [ env_julia ]

        # set dependencies as this
        env_apt_get = list(map(str.strip,env_julia)) # strip the strings
        environment_obj['env_julia'] = env_julia                


    return environment_obj      

def compute_environment_hash(env_dict):
    onv_obj_canon  = yaml.load(yaml.dump(env_dict, sort_keys=True),Loader=yaml.FullLoader)
    env_json_canon = jcs.canonicalize(onv_obj_canon)
    hash = hashlib.md5(env_json_canon).hexdigest()
    return hash[0:12]

def compute_job_hash(job_config):
    string_to_hash = ''
    if job_config.get('run_script'):
        script_args    = job_config['run_script'].split()
        scriptfile     = os.path.realpath(script_args[0])
        string_to_hash = 'script:' + scriptfile
    elif job_config.get('run_command'):
        # lets not do anything to commands ... its less known what can be in there ...
        string_to_hash = 'command:' + job_config['run_command']
    else:
        string_to_hash = 'unknown'

    # also add the upload files as part of the hash
    # this could mean the script will run with different multi inputs ... 
    if job_config.get('upload_files'):
        files = job_config['upload_files']
        if isinstance(files,str):
            files = [ files ]
        realfiles = []
        for f in sorted(files):
            realfiles.append( os.path.realpath(f) )
        string_to_hash = string_to_hash + ":" + ",".join(realfiles)

    if job_config.get('input_file'):
        inputpath = os.path.realpath(job_config.get('input_file'))
        string_to_hash = string_to_hash + ":" + inputpath

    # also add the input file which should be pointing to a specific file 
    # so if this is different for some reason (but with all the same python file, args, and upload files)
    # this means that we should probably run 

    # NO! 
    # this will be differentiated with the UID at run time
    # to mutualize more the upload files + script file ...
    # if 'input_files' in job_config and job_config['input_files'] is not None:
    #    string_to_hash = string_to_hash + ':' + job_config['input_files']

    hash = hashlib.md5(string_to_hash.encode()).hexdigest()
    return hash[0:12]

def generate_unique_id():
    return str(uuid.uuid4())

def compute_job_command(instance,script_dir,job_config):
    script_command = ''
    if 'run_script' in job_config and job_config['run_script'] is not None:
        script_args = job_config['run_script'].split()
        filename = os.path.basename(script_args[0])
        file_ext = os.path.splitext(filename)[1]
        script_args.pop(0)
        if file_ext.lower() == '.py':
            # use -u if i you want to skip stdout buffering (used by tail function)
            script_command = "python3 " + instance.path_join( script_dir , filename ) + " " + " ".join(script_args)
        elif file_ext.lower() == '.jl':
            script_command = "julia " + instance.path_join( script_dir , filename ) + " " + " ".join(script_args)
        else:
            script_command = "echo 'SCRIPT NOT HANDLED'"
    elif 'run_command' in job_config and job_config['run_command']:
        script_command = job_config['run_command']
    else:
        script_command = "echo 'NO SCRIPT DEFINED'" 

    return script_command 

def resolve_paths(instance,the_file,ref_file,remote_ref_dir,mutualize=True):

    # if we have a script
    # let's make sure we upload the right directory structure so we dont have to change the script
    if ref_file:
        ref_args        = ref_file.split()
        ref_abs_path    = os.path.abspath(ref_args[0])
        ref_abs_dir     = os.path.dirname(ref_abs_path)
        if os.path.isabs(the_file):
            local_abs_path   = the_file
        else:
            local_abs_path   = os.path.join(ref_abs_dir,the_file)
        file_abs_dir    = os.path.dirname(local_abs_path)
    else:
        ref_abs_path    = os.path.abspath(os.getcwd())
        ref_abs_dir     = ref_abs_path #os.path.dirname(ref_abs_path)
        if os.path.isabs(the_file):
            local_abs_path   = the_file
        else:
            local_abs_path   = os.path.join(ref_abs_dir,the_file)
        file_abs_dir    = os.path.dirname(local_abs_path)
    
    # this is a subdir: we have to respect the structure
    if file_abs_dir.startswith(ref_abs_dir):
        rel_path = local_abs_path.replace(ref_abs_dir,"")
        if rel_path.startswith(os.sep):
            rel_path = rel_path[1:]
        external = False
    # this is not a subdir ! we gotta create a virtual subdir that 
    else:
        #rel_path = local_abs_path.replace(os.sep,'_')
        # we gotta have to handle Windows situation here and transform the drive as a path bite
        rel_path = local_abs_path # LETS USE THE ABS PATH as a REL PATH !
        if rel_path.startswith(os.sep):
            rel_path = rel_path[1:]
        drive , path = os.path.splitdrive(rel_path)
        if drive:
            rel_path = os.path.join( drive.replace(':','') , path ) 
            if rel_path.startswith('//'): # //network_drive
                rel_path = rel_path[2:]
            elif rel_path.startswith(os.sep): # /absolute
                rel_path = rel_path[1:]
        external = True

    # use the first case if you want to leave stuff in the job's directory
    if not mutualize:
        remote_abs_path = instance.path_join(remote_ref_dir,rel_path)
        remote_rel_path = rel_path
    # use the second case if you want to mutualize uploads
    else:
        remote_abs_path = local_abs_path # always use full path in one target directory
        if remote_abs_path.startswith(os.sep):
            remote_abs_path = remote_abs_path[1:] # make relative
        remote_rel_path = remote_abs_path
        remote_abs_path = instance.path_join(remote_ref_dir,remote_abs_path)

    # local_abs_path  = the local file path
    # rel_path        = the relative path (from the job directory aka script dir)
    # remote_abs_path = the absolute remote file path
    # remote_rel_path = the relative remote path (from remote_ref_dir)
    # external        = is the file external to the script directory structure
    return local_abs_path , rel_path, remote_abs_path , remote_rel_path , external

 






import cloudsend
import asyncio
import sys 
import psutil
import subprocess
import multiprocessing 
import time
from cloudsend.maestroserver import main as server_main
from cloudsend.maestroclient import maestro_client

# if [[ $(ps aux | grep "cloudsend.maestroserver" | grep -v 'grep') ]] ; then
#     echo "CloudSend server already running"
# else
#     if [[ $(ps -ef | awk '/[m]aestroserver/{print $2}') ]] ; then
#         ps -ef | awk '/[m]aestroserver/{print $2}' | xargs kill 
#     fi
#     echo "Starting CloudSend server ..."
#     python3 -u -m cloudsend.maestroserver
# fi

def cloudsend_kill(p):
    try:
        p.kill()
    except:
        pass
    try:
        p.terminate()
    except:
        pass

def is_cloudsend_process(p,name='maestroserver'):
    try:
        if name in p.name():
            return True
        for arg in p.cmdline():
            if name in arg:
                return True
        return False
    except psutil.AccessDenied:
        return False
    except psutil.NoSuchProcess:
        return False

def cli(command):
    start_server() # locally we use python to start the server. maximizes the chance to be Windows complatible
    maestro_client(command)

def start_server():
    server_started = False
    for pid in psutil.pids():
        try:
            p = psutil.Process(pid)
            if is_cloudsend_process(p,"cloudsend.maestroserver"):
                print("[CloudSend server already started]")
                server_started = True
                break
        except psutil.NoSuchProcess:
            pass
    if not server_started:
        # make sure we kill any rogue processes
        for pid in psutil.pids():
            try:
                p = psutil.Process(pid)
                for pp in p.children(recursive=True):
                    if is_cloudsend_process(pp):
                        cloudsend_kill(pp)
                if is_cloudsend_process(p):
                    cloudsend_kill(p)
            except psutil.NoSuchProcess:
                pass
        print("[Starting CloudSend server ...]")
        subprocess.Popen(['python3','-u','-m','cloudsend.maestroserver'])
        time.sleep(1)
        #q = multiprocessing.Queue()
        #p = multiprocessing.Process(target=server_main,args=(q,))
        #p = multiprocessing.Process(target=server_main)
        #p.daemon = True
        #p.start()

def main():
    multiprocessing.set_start_method('spawn')

    if len(sys.argv)<2:
        print("python3 -m cloudsend.cli CMD [ARGS]")
        sys.exit()

    command = sys.argv[1]

    cli(command)


if __name__ == '__main__':     
    main()


import json , os , stat , yaml

try:
    os.remove('env_command.sh')
except:
    pass
try:
    os.remove('environment.yml')
except:
    pass
try:    
    os.remove('requirements.txt')
except:
    pass
try:    
    os.remove('aptget.sh')
except:
    pass
try:    
    os.remove('env_julia.jl')
except:
    pass

with open('config.json','r') as cfg_file:
    config = json.loads(cfg_file.read())

if config['command'] is not None:
    with open('env_command.sh','w') as sh_file:
        sh_file.write(config['command'])
        sh_file.close()
        st = os.stat('env_command.sh')
        os.chmod('env_command.sh', st.st_mode | stat.S_IEXEC)

if config['env_conda'] is not None:
    with open('environment.yml','w') as yml_file:
        yml_file.write(yaml.dump(config['env_conda'], sort_keys=False))
        yml_file.close()

if config['env_pypi'] is not None:
    with open('requirements.txt','w') as txt_file:
        for dep in config['env_pypi']:
            txt_file.write(dep+'\n')
        txt_file.close()

if config['env_aptget'] is not None:
    with open('aptget.sh','w') as sh_file:
        sh_file.write('#!/usr/bin/bash\n\nsudo apt-get install -y')
        for dep in config['env_aptget']:
            sh_file.write(' ' + dep)
        sh_file.close()
        st = os.stat('aptget.sh')
        os.chmod('aptget.sh', st.st_mode | stat.S_IEXEC)

if config['env_julia'] is not None:
    with open('env_julia.jl','w') as julia_file:
        #sh_file.write('#!/usr/bin/bash\n\nsudo apt-get install -y')
        julia_file.write("using Pkg\n")
        for pkg in config['env_julia']:
            julia_file.write('Pkg.add("' + pkg + '")\n')
        julia_file.close()
        st = os.stat('env_julia.jl')
        #os.chmod('env_julia.jl', st.st_mode | stat.S_IEXEC)        


import boto3
import os , sys
from .utils import *
from cloudsend.core     import CloudSendError , CloudSendInstance , CloudSendInstanceState , CloudSendPlatform
from cloudsend.core     import cs_keypairName , cs_secGroupName , cs_secGroupNameMaestro , cs_bucketName , cs_vpcName , cs_maestroRoleName , cs_maestroProfileName, cs_maestroPolicyName , init_instance_name
from cloudsend.provider import debug
from cloudsend.providerfat import CloudSendFatProvider
from cloudsend.providerlight import CloudSendLightProvider
from botocore.exceptions import ClientError
from datetime import datetime , timedelta
from botocore.config import Config
from datetime import datetime
from dateutil.relativedelta import relativedelta
import asyncio

def aws_get_session(profile_name , region ):
    if profile_name and region:
        return boto3.session.Session( profile_name = profile_name , region_name = region )
    elif profile_name:
        return boto3.session.Session( profile_name = profile_name )
    elif region:
        return boto3.session.Session( region_name = region )
    else:
        return boto3.session.Session()

# def aws_get_config(region):
#     # it's best we create a session every time wihin the threads ...
#     if region is None:
#         return Config()
#     else:
#         return Config(region_name=region)

def aws_get_region(profile_name=None):
    if profile_name:
        my_session = boto3.session.Session(profile_name=profile_name)
    else:
        my_session = boto3.session.Session()
    region = my_session.region_name     
    return region  

def aws_get_account_id(profile_name=None):
    client = boto3.client("sts")
    return client.get_caller_identity()["Account"]        

def aws_create_keypair(session,region,keypair_name,key_filename,force=False):
    debug(1,"Creating KEYPAIR ...")
    ec2_client = session.client("ec2")
    #ec2_client = boto3.client("ec2",config=aws_get_config(region))
    ec2 = session.resource('ec2')
    #ec2 = boto3.resource('ec2',config=aws_get_config(region))

    #keypair_name = cs_keypairName + '-' + region

    keypair = None
    try:
        keypair = ec2_client.create_key_pair(
            KeyName=keypair_name,
            DryRun=True,
            KeyType='rsa',
            KeyFormat='pem'
        )
    except ClientError as e: #DryRun=True >> always an "error"
        errmsg = str(e)
        
        if 'UnauthorizedOperation' in errmsg:
            debug(1,"The account is not authorized to create a keypair, please specify an existing keypair in the configuration or add Administrator privileges to the account")
            keypair = None
            #sys.exit()
            raise CloudSendError()

        elif 'DryRunOperation' in errmsg: # we are good with credentials
            try :
                if force:
                    try:
                        ec2_client.delete_key_pair(
                            KeyName=keypair_name
                        )
                    except:
                        pass
                keypair = ec2_client.create_key_pair(
                    KeyName=keypair_name,
                    DryRun=False,
                    KeyType='rsa',
                    KeyFormat='pem'
                )
                #if region is None:
                    # get the default user region so we know what were getting ... (if it changes later etc... could be a mess)
                #    my_session = boto3.session.Session()
                #    region = my_session.region_name                
                fpath   = key_filename
                pemfile = open(fpath, "w")
                pemfile.write(keypair['KeyMaterial']) # save the private key in the directory (we will use it with SSH client)
                pemfile.close()
                os.chmod(fpath, 0o600) # change permission to use with ssh (for debugging)
                debug(2,keypair)

            except ClientError as e2: # the keypair probably exists already
                errmsg2 = str(e2)
                if 'InvalidKeyPair.Duplicate' in errmsg2:
                    #keypair = ec2.KeyPair(cs_keypairName)
                    keypairs = ec2_client.describe_key_pairs( KeyNames= [keypair_name] )
                    keypair  = keypairs['KeyPairs'][0] 
                    debug(2,keypair)
                else:
                    debug(1,"An unknown error occured while retrieving the KeyPair")
                    debug(2,errmsg2)
                    #sys.exit()
                    raise CloudSendError()

        
        else:
            debug(1,"An unknown error occured while creating the KeyPair")
            debug(2,errmsg)
            #sys.exit()
            raise CloudSendError()
    
    return keypair 

def aws_find_or_create_default_vpc(session,region):
    ec2_client = session.client("ec2")
    #ec2_client = boto3.client("ec2", config=aws_get_config(region))
    vpcs = ec2_client.describe_vpcs(
        Filters=[
            {
                'Name': 'is-default',
                'Values': [
                    'true',
                ]
            },
        ]
    )
    defaultvpc = ec2_client.create_default_vpc() if len(vpcs['Vpcs'])==0 else vpcs['Vpcs'][0] 
    return defaultvpc

def aws_create_vpc(session,region,cloudId=None):
    debug(1,"Creating VPC ...")
    vpc = None
    ec2_client = session.client("ec2")
    #ec2_client = boto3.client("ec2", config=aws_get_config(region))

    if cloudId is not None:
        vpcID = cloudId
        try :
            vpcs = ec2_client.describe_vpcs(
                VpcIds=[
                    vpcID,
                ]
            )
            vpc = vpcs['Vpcs'][0]
        except ClientError as ce:
            ceMsg = str(ce)
            if 'InvalidVpcID.NotFound' in ceMsg:
                debug(1,"WARNING: using default VPC. "+vpcID+" is unavailable")
                vpc = aws_find_or_create_default_vpc(session,region) 
            else:
                debug(1,"WARNING: using default VPC. Unknown error")
                vpc = aws_find_or_create_default_vpc(session,region) 

    else:
        debug(1,"using default VPC (no VPC ID specified in config)")
        vpc = aws_find_or_create_default_vpc(session,region)
    debug(2,vpc)

    return vpc 

def aws_create_security_group(session,region,vpc):
    debug(1,"Creating SECURITY GROUP ...")
    secGroup = None
    ec2_client = session.client("ec2")
    #ec2_client = boto3.client("ec2", config=aws_get_config(region))

    secgroups = ec2_client.describe_security_groups(Filters=[
        {
            'Name': 'group-name',
            'Values': [
                cs_secGroupName,
            ]
        },
    ])
    if len(secgroups['SecurityGroups']) == 0: # we have no security group, lets create one
        debug(1,"Creating new security group")
        secGroup = ec2_client.create_security_group(
            VpcId = vpc['VpcId'] ,
            Description = 'Allow SSH connection' ,
            GroupName = cs_secGroupName 
        )

        data = ec2_client.authorize_security_group_ingress(
            GroupId=secGroup['GroupId'],
            IpPermissions=[
                {'IpProtocol': 'tcp',
                'FromPort': 80,
                'ToPort': 80,
                'IpRanges': [{'CidrIp': '0.0.0.0/0'}]},
                {'IpProtocol': 'tcp',
                'FromPort': 22,
                'ToPort': 22,
                'IpRanges': [{'CidrIp': '0.0.0.0/0'}]}
            ])
        debug(3,'Ingress Successfully Set %s' % data)

    else:
        secGroup = secgroups['SecurityGroups'][0]
    
    if secGroup is None:
        debug(1,"An unknown error occured while creating the security group")
        #sys.exit()
        raise CloudSendError()
    
    debug(2,secGroup) 

    return secGroup

def aws_add_maestro_security_group(session,instance):
    
    region = instance.get_region()
    vpcid  = instance.get_data('VpcId')

    debug(1,"Creating MAESTRO SECURITY GROUP ...")
    secGroup = None
    ec2_client = session.client("ec2")
    #ec2_client = boto3.client("ec2", config=aws_get_config(region))

    secgroups = ec2_client.describe_security_groups(Filters=[
        {
            'Name': 'group-name',
            'Values': [
                cs_secGroupNameMaestro,
            ]
        },
    ])
    if len(secgroups['SecurityGroups']) == 0: # we have no security group, lets create one
        debug(1,"Creating new security group")
        secGroup = ec2_client.create_security_group(
            VpcId = vpcid ,
            Description = 'Allow Maestro Socket connection' ,
            GroupName = cs_secGroupNameMaestro 
        )

        data = ec2_client.authorize_security_group_ingress(
            GroupId=secGroup['GroupId'],
            IpPermissions=[
                {'IpProtocol': 'tcp',
                'FromPort': 5000,
                'ToPort': 5000,
                'IpRanges': [{'CidrIp': '0.0.0.0/0'}]}
            ])
        debug(3,'Ingress Successfully Set %s' % data)

    else:
        secGroup = secgroups['SecurityGroups'][0]
    
    if secGroup is None:
        debug(1,"An unknown error occured while creating the security group")
        #sys.exit()
        raise CloudSendError()
    
    debug(2,secGroup) 

    #ec2_objs = boto3.resource('ec2',config=aws_get_config(region))
    ec2_objs = session.resource('ec2')
    instance_aws = ec2_objs.Instance(instance.get_id())
    all_sg_ids = [sg['GroupId'] for sg in instance_aws.security_groups]  # Get a list of ids of all securify groups attached to the instance
    if secGroup['GroupId'] not in all_sg_ids:                        # Check the SG to be removed is in the list
       all_sg_ids.append(secGroup['GroupId'])                        # Adds the SG from the list
       instance_aws.modify_attribute(Groups=all_sg_ids)                  # Attach the remaining SGs to the instance

    return secGroup    

# for now just return the first subnet present in the Vpc ...
def aws_create_subnet(session,region,vpc):
    debug(1,"Creating SUBNET ...")
    ec2 = session.resource('ec2')
    #ec2 = boto3.resource('ec2',config=aws_get_config(region))
    ec2_client = session.client("ec2")
    #ec2_client = boto3.client("ec2", config=aws_get_config(region))
    vpc_obj = ec2.Vpc(vpc['VpcId'])
    for subnet in vpc_obj.subnets.all():
        subnets = ec2_client.describe_subnets(SubnetIds=[subnet.id])
        subnet = subnets['Subnets'][0]
        debug(2,subnet)
        return subnet
    return None # todo: create a subnet

def aws_create_bucket(session,region):

    debug(1,"Creating BUCKET ...")
    s3_client = session.client('s3')
    #s3_client = boto3.client('s3', config=aws_get_config(region))
    s3 = session.resource('s3')
    #s3 = boto3.resource('s3', config=aws_get_config(region))

    try :

        bucket = s3_client.create_bucket(
            ACL='private',
            Bucket=cs_bucketName,
            CreateBucketConfiguration={
                'LocationConstraint': region
            },

        )

    except ClientError as e:
        errMsg = str(e)
        if 'BucketAlreadyExists' in errMsg:
            bucket = s3.Bucket(cs_bucketName)
        elif 'BucketAlreadyOwnedByYou' in errMsg:
            bucket = s3.Bucket(cs_bucketName)

    debug(2,bucket)

    return bucket 


def aws_upload_file( session , region , bucket , file_path ):
    debug(1,"uploading FILE ...")
    s3_client = session.client('s3')
    #s3_client = boto3.client('s3', config=aws_get_config(region))
    response = s3_client.upload_file( file_path, bucket['BucketName'], 'cr-run-script' )
    debug(2,response)

def aws_find_instance(session,instance_config):

    instanceName = init_instance_name(instance_config)
    region = instance_config.get('region')

    debug(1,"Searching INSTANCE",instanceName,"...")

    #ec2_client = boto3.client("ec2", config=aws_get_config(region))
    ec2_client = session.client("ec2")

    existing = ec2_client.describe_instances(
        Filters = [
            {
                'Name': 'tag:Name',
                'Values': [
                    instanceName
                ]
            },
            {
                'Name': 'instance-state-name' ,
                'Values' : [ # everything but 'terminated' and 'shutting down' ?
                    'pending' , 'running' , 'stopping' , 'stopped'
                ]
            }
        ]
    )

    if len(existing['Reservations']) > 0 and len(existing['Reservations'][0]['Instances']) >0 :
        instance_data = existing['Reservations'][0]['Instances'][0]
        debug(1,"Found existing instance !",instance_data['InstanceId'],instanceName)
        debug(3,instance_data)
        instance = CloudSendInstance( instance_config , instance_data['InstanceId'] , instance_data )
        return instance 
    
    else:

        debug(1,"not found")

        return None 


def aws_create_instance(session,instance_config,vpc,subnet,secGroup,keypair_name):

    debug(1,"Creating INSTANCE ...")

    instanceName = init_instance_name(instance_config)

    region = instance_config.get('region')

    ec2_client = session.client("ec2")
    #ec2_client = boto3.client("ec2", config=aws_get_config(region))

    #keypair_name = cs_keypairName + '-' + region

    existing = ec2_client.describe_instances(
        Filters = [
            {
                'Name': 'tag:Name',
                'Values': [
                    instanceName
                ]
            },
            {
                'Name': 'instance-state-name' ,
                'Values' : [ # everything but 'terminated' and 'shutting down' ?
                    'pending' , 'running' , 'stopping' , 'stopped'
                ]
            }
        ]
    )

    created = False

    if len(existing['Reservations']) > 0 and len(existing['Reservations'][0]['Instances']) >0 :
        instance_data = existing['Reservations'][0]['Instances'][0]
        debug(1,"Found existing instance !",instance_data['InstanceId'])
        debug(3,instance_data)
        instance = CloudSendInstance( instance_config, instance_data['InstanceId'] , instance_data )
        return instance , created

    if instance_config.get('cpus') is not None:
        cpus_spec = {
            'CoreCount': instance_config['cpus'],
            'ThreadsPerCore': 1
        }
    else:
        cpus_spec = { }  

    if instance_config.get('gpu'):
        gpu_spec = [ { 'Type' : instance_config['gpu'] } ]
    else:
        gpu_spec = [ ]
    
    if instance_config.get('eco') :
        market_options = {
            'MarketType': 'spot',
            'SpotOptions': {
                'SpotInstanceType': 'one-time', #one-time'|'persistent',
                'InstanceInterruptionBehavior': 'terminate' #'hibernate'|'stop'|'terminate'
            }
        }     
        if 'eco_life' in instance_config and isinstance(instance_config['eco_life'],timedelta):
            validuntil = datetime.today() + instance_config['eco_life']
            market_options['SpotOptions']['ValidUntil'] = validuntil
        if 'max_bid' in instance_config and isinstance(instance_config['max_bid'],str):
            market_options['SpotOptions']['MaxPrice'] = instance_config['max_bid']
    else:
        market_options = { } 

    if instance_config.get('disk_size'):
        #TODO: improve selection of disk_type
        disk_size = instance_config['disk_size']
        if disk_size < 1024:
            volume_type = 'standard'
        else:
            volume_type = 'st1' # sc1, io1, io2, gp2, gp3
        if 'disk_type' in instance_config and instance_config['disk_type']:
            volume_type = instance_config['disk_type']
        block_device_mapping = [
            {
                'DeviceName' : '/dev/sda' ,
                'Ebs' : {
                    "DeleteOnTermination" : True ,
                    "VolumeSize": disk_size ,
                    "VolumeType": volume_type
                }
            }
        ]
    else:
        block_device_mapping = [ ]

    try:
        instances = ec2_client.run_instances(
                ImageId = instance_config['img_id'],
                MinCount = 1,
                MaxCount = 1,
                InstanceType = instance_config['type'],
                KeyName = keypair_name,
                SecurityGroupIds=[secGroup['GroupId']],
                SubnetId = subnet['SubnetId'],
                TagSpecifications=[
                    {
                        'ResourceType': 'instance',
                        'Tags': [
                            {
                                'Key': 'Name',
                                'Value': instanceName
                            },
                        ]
                    },
                ],
                ElasticGpuSpecification = gpu_spec ,
                InstanceMarketOptions = market_options,
                CpuOptions = cpus_spec,
                HibernationOptions={
                    'Configured': False
                },            
        )    
    except ClientError as ce:
        errmsg = str(ce)
        if 'InsufficientInstanceCapacity' in errmsg: # Amazon doesnt have enough resources at the moment
            debug(1,"AWS doesnt have enough SPOT resources at the moment, retrying in 2 minutes ...",errmsg)
        elif 'InvalidAMIID.Malformed' in errmsg:
            debug(1,errmsg,color=bcolors.FAIL)
            images = ec2_client.describe_images(Filters=[{'Name':'name','Values':['*Ubuntu*']}]) #Owners=['self'])
            for image in images['Images']:
                debug(1,"{0} - {1}".format(image['ImageId'],image['Name']))
            sys.exit()
        elif 'InvalidAMIID.NotFound' in errmsg:
            debug(1,errmsg,color=bcolors.FAIL)
            images = ec2_client.describe_images(Filters=[{'Name':'name','Values':['*Ubuntu*']}]) #Owners=['self'])
            for image in images['Images']:
                debug(1,"{0} - {1}".format(image['ImageId'],image['Name']))
            sys.exit()
        else:
            debug(1,"An error occured while trying to create this instance",errmsg)
        raise CloudSendError()


    created = True

    debug(2,instances["Instances"][0])

    instance = CloudSendInstance( instance_config, instances["Instances"][0]["InstanceId"] , instances["Instances"][0] )

    return instance , created

def aws_create_instance_objects(session,instance_config,keypair_name,key_filename):
    region   = instance_config.get('region')
    keypair  = aws_create_keypair(session,region,keypair_name,key_filename)
    vpc      = aws_create_vpc(session,region,instance_config.get('cloud_id')) 
    secGroup = aws_create_security_group(session,region,vpc)
    subnet   = aws_create_subnet(session,region,vpc) 
    # this is where all the instance_config is actually used
    instance , created = aws_create_instance(session,instance_config,vpc,subnet,secGroup,keypair_name)

    return instance , created 


def aws_start_instance(session,instance):
    region  = instance.get_region()
    ec2_client = session.client("ec2")
    #ec2_client = boto3.client("ec2", config=aws_get_config(instance.get_region()))

    try:
        ec2_client.start_instances(InstanceIds=[instance.get_id()])
    except ClientError as botoerr:
        errmsg = str(botoerr)
        if 'IncorrectSpotRequestState' in errmsg:
            debug(1,"Could not start because it is a SPOT instance, waiting on SPOT ...",errmsg)
            # try reboot
            try:
                ec2_client.reboot_instances(InstanceIds=[instance.get_id()])
            except ClientError as botoerr2:
                errmsg2 = str(botoerr2)
                if 'IncorrectSpotRequestState' in errmsg2:
                    debug(1,"Could not reboot instance because of SPOT ... ",errmsg2)
                    raise CloudSendError()
                    # terminate_instance(config,instance)
                
        else:
            debug(2,botoerr)
            raise CloudSendError()

def aws_stop_instance(session,instance):
    region  = instance.get_region()
    ec2_client = session.client("ec2")
    #ec2_client = boto3.client("ec2", config=aws_get_config(instance.get_region()))

    ec2_client.stop_instances(InstanceIds=[instance.get_id()])

def aws_terminate_instance(session,instance):
    region  = instance.get_region()
    ec2_client = session.client("ec2")
    #ec2_client = boto3.client("ec2", config=aws_get_config(instance.get_region()))

    ec2_client.terminate_instances(InstanceIds=[instance.get_id()])

    if instance.get_data('SpotInstanceRequestId'):

        ec2_client.cancel_spot_instance_requests(SpotInstanceRequestIds=[instance.get_data('SpotInstanceRequestId')]) 

def aws_reboot_instance(session,instance):
    region  = instance.get_region()
    ec2_client = session.client("ec2")
    #ec2_client = boto3.client("ec2", config=aws_get_config(instance.get_region()))

    ec2_client.reboot_instances(InstanceIds=[instance.get_id()],)


def aws_update_instance_info(session,instance):
    region = instance.get_region()
    #ec2_client   = boto3.client("ec2", config=aws_get_config(region))
    ec2_client = session.client("ec2")
    instances    = ec2_client.describe_instances( InstanceIds=[instance.get_id()] )
    instance_new_data = instances['Reservations'][0]['Instances'][0]

    debug(3,instance_new_data)

    #instance = CloudSendInstance( region , instance.get_name() , instance.get_id() , instance_config, instance_new )
    # proprietary values
    instance.set_dns_addr(instance_new_data.get('PublicDnsName'))
    instance.set_ip_addr(instance_new_data.get('PublicIpAddress'))
    instance.set_dns_addr_priv(instance_new_data.get('PrivateDnsName'))
    instance.set_ip_addr_priv(instance_new_data.get('PrivateIpAddress'))
    statestr = instance_new_data.get('State').get('Name').lower()
    #'terminated' | 'shutting-down' | 'pending' | 'running' | 'stopping' | 'stopped'
    state = CloudSendInstanceState.UNKNOWN
    if statestr == "pending":
        state = CloudSendInstanceState.STARTING
    elif statestr == "running":
        state = CloudSendInstanceState.RUNNING
    elif statestr == 'stopping':
        state = CloudSendInstanceState.STOPPING
    elif statestr == 'stopped':
        state = CloudSendInstanceState.STOPPED
    elif statestr == 'shutting-down':
        state = CloudSendInstanceState.TERMINATING
    elif statestr == 'terminated':
        state = CloudSendInstanceState.TERMINATED
    instance.set_state(state)
    instance.set_data(instance_new_data)
    instance.set_reachability(False)

    # check further status
    if state == CloudSendInstanceState.RUNNING:
        status_res = ec2_client.describe_instance_status( InstanceIds=[instance.get_id()] )
        status_res = status_res['InstanceStatuses'][0]
        # if status_res['InstanceStatus']['Status'] == 'ok': #and status_res['SystemStatus']['Status'] == 'ok':
        #     instance.set_reachability(True)
        # this seems okay to force-try to connect...
        instance.set_reachability(True)

    platform_details = instance_new_data.get('PlatformDetails').lower()
    if 'linux' in platform_details:
        instance.set_platform(CloudSendPlatform.LINUX)
    elif 'windows' in platform_details:
        instance.set_platform(CloudSendPlatform.WINDOWS_WSL)

    return instance


def aws_grant_admin_rights(session,instance):
    region = instance.get_region()
    iam_client = session.client('iam')
    #iam_client = boto3.client('iam', config=aws_get_config(region))
    id_client = session.client('sts')
    #id_client  = boto3.client("sts", config=aws_get_config(region))
    ec2_client = session.client('ec2')
    #ec2_client = boto3.client("ec2", config=aws_get_config(region))
    
    account_id = id_client.get_caller_identity()["Account"]
    #session    = boto3.session.Session()
    #region_flt = session.region_name if not region else region

    #Following trust relationship policy can be used to provide access to assume this role by a particular AWS service in the same account
    trust_relationship_policy_another_aws_service = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Principal": {
                    "Service": "ec2.amazonaws.com"
                },
                "Action": "sts:AssumeRole"
            }
        ]
    }

    try:
        create_role_res = iam_client.create_role(
            RoleName=cs_maestroRoleName,
            AssumeRolePolicyDocument=json.dumps(trust_relationship_policy_another_aws_service),
            Description='CloudSend Admin Role'
        )
    except ClientError as error:
        if error.response['Error']['Code'] == 'EntityAlreadyExists':
            debug(2,'Role already exists ...')
        else:
            debug(1,'Unexpected error occurred... Role could not be created', error)
            return
            
    policy_json = {
        "Version": "2012-10-17",
        "Statement": [{
            "Effect": "Allow",
            "Action": [
                "ec2:*"
            ],
            "Resource": "*",
            # "Condition": {
            #     "StringEquals": {
            #          "ec2:Region": region_flt
            #     }
            # }
        }]
    }

    policy_arn = ''

    try:
        policy_res = iam_client.create_policy(
            PolicyName=cs_maestroPolicyName,
            PolicyDocument=json.dumps(policy_json)
        )
        policy_arn = policy_res['Policy']['Arn']
    except ClientError as error:
        if error.response['Error']['Code'] == 'EntityAlreadyExists':
            debug(2,'Policy already exists... hence using the same policy')
            policy_arn = 'arn:aws:iam::' + account_id + ':policy/' + cs_maestroPolicyName
        else:
            debug(1,'Unexpected error occurred... hence cleaning up', error)
            iam_client.delete_role(
                RoleName= cs_maestroRoleName
            )
            debug(1,'Role could not be created...', error)
            return

    try:
        policy_attach_res = iam_client.attach_role_policy(
            RoleName=cs_maestroRoleName,
            PolicyArn=policy_arn
        )
    except ClientError as error:
        debug(1,'Unexpected error occurred... hence cleaning up')
        iam_client.delete_role(
            RoleName= cs_maestroRoleName
        )
        debug(1,'Role could not be created...', error)
        return

    debug(2,'Role {0} successfully got created'.format(cs_maestroRoleName))


    profile_arn = ''
    try:
        profile_response = iam_client.create_instance_profile(InstanceProfileName=cs_maestroProfileName)
        profile_arn = profile_response['InstanceProfile']['Arn']
    except ClientError as error:
        if error.response['Error']['Code'] == 'EntityAlreadyExists':
            debug(2,'Instance profile already exists ...')
            profile_response = iam_client.get_instance_profile(InstanceProfileName=cs_maestroProfileName)
            profile_arn = profile_response['InstanceProfile']['Arn']
    try:
        iam_client.add_role_to_instance_profile(InstanceProfileName=cs_maestroProfileName,RoleName=cs_maestroRoleName)
    except ClientError as error:
        if error.response['Error']['Code'] == 'LimitExceeded':
            debug(2,'Role has already been added ...')
        else:
            debug(1,error)
            return

    try:
        response = ec2_client.associate_iam_instance_profile(
            IamInstanceProfile={
                'Arn': profile_arn,
                'Name': cs_maestroProfileName
            },
            InstanceId=instance.get_id()
        )
    except ClientError as error:
        if error.response['Error']['Code'] == 'IncorrectState':
            debug(2,'Instance Profile already associated ...')
        else:
            debug(1,error)

    debug(1,"MAESTRO role added to instance",instance.get_id(),instance.get_name())


def aws_setup_auto_stop(session,instance):
    ec2_client = session.client("ec2")
    ec2_client.modify_instance_attribute(
        InstanceId=instance.get_id(),
        InstanceInitiatedShutdownBehavior={
            'Value': 'stop'
        }
    )

def aws_get_suggested_image(session,region):
    ec2_client = session.client("ec2")

    debug(2,'Getting suggested image ID')

    flt_creation = []
    for month in range(1):
        date_x_months_ago = datetime.now() - relativedelta(months=month)
        flt = str(date_x_months_ago.year) + '-' + str(date_x_months_ago.month) + '*'
        flt_creation.append(flt)
    

    results = ec2_client.describe_images(
        Filters=
        [
            {'Name':'name','Values':['AWS Deep Learning*Ubuntu*']} ,
#            {'Name':'is-public','Values':['true']} ,
#            {'Name':'architecture','Values':['x86_64']},
            {'Name':'creation-date','Values':flt_creation}
        ],
        Owners=['self','amazon']
    )
    images = results['Images']

    images = sorted(images, key=lambda d: d['Name']) 

    for image in images:
        debug(2,image['Name'],image['Description'])
    
    images = sorted(images, key=lambda d: d['CreationDate']) 

    if images and len(images)>0:
        return images[len(images)-1]['ImageId'] , 'ubuntu' , 't2.micro' # nano: pip gets killed :(

##########
# PUBLIC #
##########

class AWSCloudSendProviderImpl():

    def find_instance(self,config):
        session = self.get_session(config)
        return aws_find_instance(session,config)

    def update_instance_info(self,instance):
        session = self.get_session(instance)
        aws_update_instance_info(session,instance)

    def create_keypair(self,region,force=False):
        keypair_name = self.get_keypair_name(self._profile_name,region)
        key_filename = self.get_key_filename(self._profile_name,region)
        session = self.get_session(config)
        aws_create_keypair(session,region,keypair_name,key_filename,force)

    def create_instance_objects(self,config):
        keypair_name = self.get_keypair_name(self._profile_name,config.get('region'))
        key_filename = self.get_key_filename(self._profile_name,config.get('region'))
        session = self.get_session(config)
        return aws_create_instance_objects(session,config,keypair_name,key_filename)

    def start_instance(self,instance):
        session = self.get_session(instance)
        aws_start_instance(session,instance)

    def stop_instance(self,instance):
        session = self.get_session(instance)
        aws_stop_instance(session,instance)

    def terminate_instance(self,instance):
        session = self.get_session(instance)
        aws_terminate_instance(session,instance)

    def reboot_instance(self,instance):
        session = self.get_session(instance)
        aws_reboot_instance(session,instance)

    def get_region(self):
        return aws_get_region(self._profile_name)
    
    def get_account_id(self):
        return aws_get_account_id(self._profile_name)

    def set_profile(self,profile_name):
        boto3.setup_default_session(profile_name=profile_name)

    def get_suggested_image(self,region):
        session = self.get_session(region)
        return aws_get_suggested_image(session,region)

    def get_session(self,obj):
        if isinstance(obj,dict):
            region = obj.get('region')
        elif isinstance(obj,str):
            region = obj
        else:
            region = obj.get_region()
        return aws_get_session(self._profile_name,region)        


class AWSCloudSendFatProvider(CloudSendFatProvider,AWSCloudSendProviderImpl):

    def __init__(self, conf):
        CloudSendFatProvider.__init__(self,conf)

    def find_instance(self,config):
        return AWSCloudSendProviderImpl.find_instance(self,config)

    def update_instance_info(self,instance):
        AWSCloudSendProviderImpl.update_instance_info(self,instance)

    def create_keypair(self,region,force=False):
        AWSCloudSendProviderImpl.create_keypair(self,region,force=False)

    def create_instance_objects(self,config):
        return AWSCloudSendProviderImpl.create_instance_objects(self,config)

    def start_instance(self,instance):
        AWSCloudSendProviderImpl.start_instance(self,instance)

    def stop_instance(self,instance):
        AWSCloudSendProviderImpl.stop_instance(self,instance)

    def terminate_instance(self,instance):
        AWSCloudSendProviderImpl.terminate_instance(self,instance)

    def reboot_instance(self,instance):
        AWSCloudSendProviderImpl.reboot_instance(self,instance)

    def get_region(self):
        return AWSCloudSendProviderImpl.get_region(self)
    
    def get_account_id(self):
        return AWSCloudSendProviderImpl.get_account_id(self)

    def set_profile(self,profile_name):
        AWSCloudSendProviderImpl.set_profile(self,profile_name)     

    def get_suggested_image(self,region):
        return AWSCloudSendProviderImpl.get_suggested_image(self,region)

    def get_recommended_cpus(self,inst_cfg):
        return self._get_instancetypes_attribute(inst_cfg,"instancetypes-aws.csv","Instance type","Valid cores",list)

    def get_cpus_cores(self,inst_cfg):
        return self._get_instancetypes_attribute(inst_cfg,"instancetypes-aws.csv","Instance type","Cores",int)


class AWSCloudSendLightProvider(CloudSendLightProvider,AWSCloudSendProviderImpl):        

    def __init__(self, conf):
        CloudSendLightProvider.__init__(self,conf)

    def find_instance(self,config):
        return AWSCloudSendProviderImpl.find_instance(self,config)

    def update_instance_info(self,instance):
        AWSCloudSendProviderImpl.update_instance_info(self,instance)

    def create_keypair(self,region,force=False):
        AWSCloudSendProviderImpl.create_keypair(self,region,force=False)

    def create_instance_objects(self,config):
        return AWSCloudSendProviderImpl.create_instance_objects(self,config)

    def start_instance(self,instance):
        AWSCloudSendProviderImpl.start_instance(self,instance)

    def stop_instance(self,instance):
        AWSCloudSendProviderImpl.stop_instance(self,instance)

    def terminate_instance(self,instance):
        AWSCloudSendProviderImpl.terminate_instance(self,instance)

    def reboot_instance(self,instance):
        AWSCloudSendProviderImpl.reboot_instance(self,instance)

    def get_region(self):
        return AWSCloudSendProviderImpl.get_region(self)
    
    def get_account_id(self):
        return AWSCloudSendProviderImpl.get_account_id(self)

    def set_profile(self,profile_name):
        AWSCloudSendProviderImpl.set_profile(self,profile_name)   

    def get_suggested_image(self,region):
        return AWSCloudSendProviderImpl.get_suggested_image(self,region)

    def grant_admin_rights(self,instance):
        session = self.get_session(instance)
        aws_grant_admin_rights(session,instance)   

    def add_maestro_security_group(self,instance):
        session = self.get_session(instance)
        aws_add_maestro_security_group(session,instance)

    def setup_auto_stop(self,instance):
        session = self.get_session(instance)
        aws_setup_auto_stop(session,instance)


from abc import ABC , abstractmethod
import cloudsend.utils as cloudsendutils
import sys , json , os , time 
import re
from io import BytesIO
import csv , io
import pkg_resources
from cloudsend.core import *
from enum import IntFlag
#import multiprocessing
import asyncio
import asyncssh


COMMAND_ARGS_SEP = '__:__'
ARGS_SEP         = '__,__'
PROVIDER_CONFIG  = 'state.config.json'

class CloudSendProvider(ABC):

    def __init__(self, conf=None):

        if conf is None:
            conf = dict()

        self._init(conf)
        #CloudSendProvider._init(self,conf)

        #self._instances_locks = dict()
        #self._provider_lock   = multiprocessing.Manager().Lock()
        #self._thread_safe_ultra = True # set to False if you want to try per instance locks

    def _init(self,conf):
        self._state = CloudSendProviderState.NEW

        self.DBG_LVL = conf.get('debug',1)
        self.DBG_PREFIX = None
        global DBG_LVL
        DBG_LVL = conf.get('debug',1)

        self._config  = conf
        self._auto_stop = conf.get('auto_stop',False)

        self._profile_name = self._config.get('profile')
        if self._config.get('profile'):
            self.set_profile(self._config.get('profile'))   
        
        self._save_config()

    def debug_set_prefix(self,value):
        self.DBG_PREFIX = value
        global DBG_PREFIX
        DBG_PREFIX = value

    def debug(self,level,*args,**kwargs):
        if level <= self.DBG_LVL:
            if 'color' in kwargs:
                color = kwargs['color']
                listargs = list(args)
                #listargs.insert(0,color)
                if isinstance(listargs[0], (bytes, bytearray)):
                    str_concat = listargs[0].decode()
                else:
                    str_concat = str(listargs[0])
                listargs[0] = color + str_concat
                listargs.append(bcolors.ENDC)
                args = tuple(listargs)
                kwargs.pop('color')
            try:
                if not sys.stdout: # or sys.stdout.closed==True:
                    # sys.stdout = sys.__stdout__
                    # print("SYS.STDOUT is DEAD")
                    # sys.stdout = None
                    return

                if self.DBG_PREFIX:
                    print(self.DBG_PREFIX,*args,**kwargs) 
                else:
                    print(*args,**kwargs) 
                sys.stdout.flush()
                sys.stderr.flush()
            except:
                pass

    def get_state(self):
        return self._state

    def _get_or_create_instance(self,instance):

        inst_cfg = instance.get_config_DIRTY()

        # this is important because we can very well have watch() daemon and wait() method run at the same time
        # and both those methods may call '_start_and_update_instance' > '_get_or_create_instance'
        #with self._lock(instance):
            
        instance = self.find_instance(inst_cfg)

        if instance is None:
            instance , created = self.create_instance_objects(inst_cfg)
        else:
            self.update_instance_info(instance) # make sure we update
            created = False

        return instance , created  

    def _start_and_update_instance(self,instance):

        try:
            # CHECK EVERY TIME !
            new_instance , created = self._get_or_create_instance(instance)

            old_id = instance.get_id()
            
            # make sure we update the instance with the new instance data
            instance.update_from_instance(new_instance)

            if instance.get_id() != old_id:
                self.debug(2,"Instance has changed",old_id,"VS",instance.get_id())
                #self._instances_states[instance.get_id()] = { 'changed' : True }
                if not hasattr(self,'_instances_states'):
                    self._instances_states = dict()
                self._instances_states[instance.get_name()] = { 'changed' : True }

        except CloudSendError as cre:

            instance.set_invalid(True)

            self.debug(1,cre)
        
        except Exception as e:
            self.debug(1,e)

            raise e

    def _resolve_dpl_job_paths(self,upfile,dpl_job):
        if self._mutualize_uploads:
            instance = dpl_job.get_instance()
            remote_ref_dir = instance.path_join( instance.get_home_dir() , 'run' , 'files')
            return cloudsendutils.resolve_paths(instance,upfile,dpl_job.get_config('run_script'),remote_ref_dir,True) 
        else:
            instance = dpl_job.get_instance()
            return cloudsendutils.resolve_paths(instance,upfile,dpl_job.get_config('run_script'),dpl_job.get_path(),False)

    # def _lock(self,instance):
    #     if not self._thread_safe_ultra:
    #         if not instance.get_name() in self._instances_locks:
    #             self._instances_locks[instance.get_name()] = multiprocessing.Manager().Lock()
    #         return self._instances_locks[instance.get_name()]
    #     else:
    #         return self._provider_lock


    async def _wait_and_connect(self,instance):

        # wait for instance to be ready 
        await self._wait_for_instance(instance)

        # connect to instance
        ssh_conn = await self._connect_to_instance(instance)
        if ssh_conn is not None:
            ftp_client = await ssh_conn.start_sftp_client()

            return instance.get_id() , ssh_conn , ftp_client        
        else:
            return instance.get_id() , None , None

    async def _wait_for_instance(self,instance):
        
        # get the public DNS info when instance actually started (todo: check actual state)
        waitFor = True
        while waitFor:
            # we may be heavily updating the instace already ...
            # with self._lock(instance):
            self.update_instance_info(instance)

            self.serialize_state()

            lookForDNS       = instance.get_dns_addr() is None
            lookForIP        = instance.get_ip_addr() is None
            instanceState    = instance.get_state()
            reachability     = instance.get_reachability()

            lookForState = True
            # 'pending'|'running'|'shutting-down'|'terminated'|'stopping'|'stopped'
            if instanceState == CloudSendInstanceState.STOPPED:
                try:
                    # restart the instance
                    self.start_instance(instance)
                except CloudSendError:
                    await self.hard_reset_instance(instance)


            elif instanceState == CloudSendInstanceState.RUNNING:
                lookForState = False

            elif instanceState == CloudSendInstanceState.TERMINATING or instanceState == CloudSendInstanceState.TERMINATED:
                try:
                    self._start_and_update_instance(instance)
                except:
                    pass

            waitFor = lookForDNS or lookForState or not reachability
            if waitFor:
                if lookForDNS:
                    debug(1,"waiting for",instance.get_name(),"...",instanceState.name)
                else:
                    if lookForIP:
                        debug(1,"waiting for",instance.get_name(),"...",instanceState.name)
                    elif lookForState:
                        debug(1,"waiting for",instance.get_name(),"...",instanceState.name," IP =",instance.get_ip_addr())
                    elif not reachability:
                        debug(1,"waiting for",instance.get_name(),"...",instanceState.name," IP =",instance.get_ip_addr(),"(waiting to be reachable)")
                    else:
                        debug(1,"waiting for",instance.get_name(),"...",instanceState.name," IP =",instance.get_ip_addr())
                 
                await asyncio.sleep(10)

        self.debug(2,instance)     

    # async because the one of the fat client will be async ...
    async def hard_reset_instance(self,instance):
        self.terminate_instance(instance)
        try :
            self._start_and_update_instance(instance)
        except:
            pass

    async def _connect_to_instance(self,instance,**kwargs):
        # ssh into instance and run the script 
        region = instance.get_region()
        #if region is None:
        #    region = self.get_user_region(self._config.get('profile'))
        keypair_filename = self.get_key_filename(self._config.get('profile'),region)
        if not os.path.exists(keypair_filename):
            self.create_keypair(region,True)
        k = asyncssh.read_private_key(keypair_filename)
        self.debug(1,"connecting to",instance.get_name(),"@",instance.get_ip_addr())
        retrys = 0 
        while True:
            try:
                conn = await asyncssh.connect(host=instance.get_dns_addr(),username=instance.get_config('img_username'),client_keys=[k],known_hosts=None,**kwargs) #,password=’mypassword’)
                break
            except Exception as cexc:
                # check whats going on
                self.update_instance_info(instance)
                if instance.get_state() != CloudSendInstanceState.RUNNING:
                    await self._wait_for_instance(instance)
                if retrys < 5:
                    self.debug(1,cexc)
                    await asyncio.sleep(4)
                    self.debug(1,"Retrying ...")
                    retrys = retrys + 1
                elif retrys == 5 and instance.get_state() == CloudSendInstanceState.RUNNING:
                    retrys = retrys + 1
                    self.debug(1,"Trying a reboot ...")
                    self.reboot_instance(instance)
                    await asyncio.sleep(4)
                    await self._wait_for_instance(instance)
                elif retrys == 6 and instance.get_state() == CloudSendInstanceState.RUNNING:
                    retrys = retrys + 1
                    self.debug(1,"Trying a hard reset ...")
                    await self.hard_reset_instance(instance)
                    await asyncio.sleep(4)
                    await self._wait_for_instance(instance)
                else:
                    self.debug(1,cexc)
                    self.debug(0,"ERROR! instance is unreachable: ",instance,color=bcolors.FAIL)
                    return None
            # except OSError as ose:
            #     if retrys < 5:
            #         self.debug(1,ose)
            #         await asyncio.sleep(4)
            #         self.debug(1,"Retrying (2) ...")
            #         retrys = retrys + 1
            #     else:
            #         self.debug(1,cexc)
            #         self.debug(0,"ERROR! instance is unreachable: ",instance)
            #         return None

        self.debug(1,"connected to ",instance.get_ip_addr())    

        return conn              


    async def _exec_command(self,ssh_conn,command):
        self.debug(2,"Executing ",format( command ))
        try:
            proc = await ssh_conn.create_process(command)
            return proc.stdout , proc.stderr 
        except (OSError, asyncssh.Error) as sshe:
            self.debug(1,'SSH connection failed: ' + str(sshe),color=bcolors.FAIL)
            #self.debug(1,"The SSH Client has been disconnected!")
            self.debug(2,sshe)
            raise sshe
            
    async def _run_ssh_commands(self,instance,ssh_conn,commands):
        for command in commands:
            self.debug(2,"Executing ",format( command['cmd'] ),"output",command['out'])
            try:
                if command['out']:
                    async with ssh_conn.create_process(command['cmd']) as proc:
                        # for l in line_buffered(proc.stdout):
                        #     self.debug(1,l,end='')
                        async for line in proc.stdout:
                            if not line:
                                break
                            self.debug(1,line,end='')

                        errmsg = await proc.stderr.read()
                        dbglvl = 1 if errmsg else 2
                        self.debug(dbglvl,"Errors")
                        self.debug(dbglvl,errmsg)

                else:
                    if not 'output' in command:
                        output = instance.path_join( instance.get_home_dir() , 'run' , 'out.log' )
                    else:
                        output = command['output']
                    await ssh_conn.create_process(command['cmd'] + " 1>"+output+" 2>&1 &")
            except (OSError, asyncssh.Error) as sshe:
                self.debug(1,'SSH connection failed: ' + str(sshe),color=bcolors.FAIL)
                #self.debug(1,"The SSH Client has been disconnected!")
                self.debug(2,sshe)
                raise sshe

    async def sftp_put_remote_file(self,ftp_client,name):
        ofile = await ftp_client.open(name,'w')
        await ofile.write(self._get_remote_file(name))
        await ofile.close()

    async def sftp_put_string(self,ftp_client,name,string):
        ofile = await ftp_client.open(name,'w')
        await ofile.write(string)
        await ofile.close()

    async def sftp_put_bytes(self,ftp_client,name,bytes_,encoding='utf-8'):
        #byte_str = bytes_io.read()
        # Convert to a "unicode" object
        string = bytes_.decode(encoding)
        ofile = await ftp_client.open(name,'w',encoding=encoding)
        await ofile.write(string)
        await ofile.close()

    async def _test_reupload(self,instance,file_test,ssh_conn,isfile=True):
        re_upload = False
        if isfile:
            stdout0, stderr0 = await self._exec_command(ssh_conn,"[ -f "+file_test+" ] && echo \"ok\" || echo \"not_ok\";")
        else:
            stdout0, stderr0 = await self._exec_command(ssh_conn,"[ -d "+file_test+" ] && echo \"ok\" || echo \"not_ok\";")
        result = await stdout0.read()
        if "not_ok" in result: #result.decode():
            debug(1,"re-upload of files ...")
            re_upload = True
        return re_upload                

    def _get_instancetypes_attribute(self,inst_cfg,resource_file,type_col,attr,return_type):

        # Could be any dot-separated package or module name or a "Requirement"
        resource_package = 'cloudsend'
        resource_path = os.sep.join(('resources', resource_file))  # Do not use os.path.join()
        #template = pkg_resources.resource_string(resource_package, resource_path)
        # or for a file-like stream:
        #template = pkg_resources.resource_stream(resource_package, resource_path)        
        #with open('instancetypes-aws.csv', newline='') as csvfile:
        csvstr = pkg_resources.resource_string(resource_package, resource_path)
        self._csv_reader = csv.DictReader(io.StringIO(csvstr.decode()))
        for row in self._csv_reader:
            if row[type_col] == inst_cfg.get('type'):
                if return_type == list:
                    arr = row[attr].split(',')
                    res = [ ]
                    for x in arr:
                        try:
                            res.append(int(x))
                        except: 
                            pass
                    if len(res)==0:
                        return None
                    return res
                elif return_type == int:
                    try:
                        res = int(row[attr])
                    except:
                        return None
                    return res 
                elif return_type == str:
                    return row[attr]
                else:
                    return raw[attr]
        return         

    def _get_resource_file(self,resource_file):
        resource_package = 'cloudsend'
        resource_path = os.sep.join(('resources', resource_file))  # Do not use os.path.join()
        #return pkg_resources.resource_stream(resource_package, resource_path)  
        return pkg_resources.resource_string(resource_package, resource_path).decode()

    def _get_remote_file(self,file):
        return self._get_resource_file(os.sep.join(('remote_files',file)))

    def get_key_filename(self,profile_name,region):
        userid = profile_name
        if not userid:
            userid = 'default'
        #key_filename = 'cloudsend-'+str(userid)+'-'+str(region)+'.pem'
        #key_filename = 'cloudsend-'+str(region)+'.pem'
        key_filename = 'cloudsend-'+self.get_account_id()+'-'+str(region)+'.pem'
        return key_filename

    def get_keypair_name(self,profile_name,region):
        userid = profile_name
        if not userid:
            userid = 'default'
        #key_filename = cs_keypairName+'-'+str(userid)+'-'+str(region)
        key_filename = cs_keypairName+'-'+str(region)
        return key_filename

    @abstractmethod
    def get_region(self):
        pass

    @abstractmethod
    def get_account_id(self):
        pass

    # @abstractmethod
    # def get_user_id(self):
    #     pass

    @abstractmethod
    def set_profile(self,profile_name):
        pass

    @abstractmethod
    def create_keypair(self,region,force=False):
        pass

    @abstractmethod
    def create_instance_objects(self,config):
        pass

    @abstractmethod
    def find_instance(self,config):
        pass

    @abstractmethod
    def start_instance(self,instance):
        pass

    @abstractmethod
    def stop_instance(self,instance):
        pass

    @abstractmethod
    def terminate_instance(self,instance):
        pass

    @abstractmethod
    def reboot_instance(self,instance):
        pass

    @abstractmethod
    async def reset_instance(self,instance):
        pass

    @abstractmethod
    def update_instance_info(self,instance):
        pass

    @abstractmethod
    async def wakeup(self):
        pass

    @abstractmethod
    async def start(self,reset=False):
        pass

    def _save_config(self):
        with open(PROVIDER_CONFIG,'w') as config_file:
            config_file.write( json.dumps(self._config,indent=4) )

    def _cfg_add_objects(self,key_name,config_list_obj,save=True):
        # config could be a file path
        if isinstance(config_list_obj,str) and os.path.isfile(config_list_obj):
            config = get_config(config_list_obj)
        # config could be a json string stream
        elif isinstance(config_list_obj,str):
            try:
                config = json.loads(config_list_obj)
            except:
                config = dict()
        # config could be an array (of instances, envs or jobs specs)
        elif isinstance(config_list_obj,list):
            config = { key_name : config_list_obj }
        # config could be a cloudsend-compatible config
        elif isinstance(config_list_obj,dict) and key_name in config_list_obj:
            config = config_list_obj
        # config could be a singleton instance,env or job config
        elif isinstance(config_list_obj,dict):
            config = { key_name : [ config_list_obj ] }
        # other cases
        else:
            config = dict()

        self.debug(2,"ADDING objects",config)
        
        if key_name not in config:
            self.debug(1,"No "+key_name+" specified in config file",config)
            return
        
        if key_name not in self._config:
            self._config[key_name] = []

        for cfg in config[key_name]:
            cfg[K_LOADED] =False # make sure we mark it as not-loaded
            self._config[key_name].append( cfg )

        # no need, this only plays with global vars...
        # self._init(self._config)
        if save:
            self._save_config()

    async def cfg_add_instances(self,config):
        CloudSendProvider._cfg_add_objects(self,'instances',config)

    async def cfg_add_environments(self,config):
        CloudSendProvider._cfg_add_objects(self,'environments',config)

    async def cfg_add_jobs(self,config):
        CloudSendProvider._cfg_add_objects(self,'jobs',config)

    async def cfg_add_config(self,config):
        CloudSendProvider._cfg_add_objects(self,'instances',config,False)
        CloudSendProvider._cfg_add_objects(self,'environments',config,False)
        CloudSendProvider._cfg_add_objects(self,'jobs',config,False)
        self._save_config()

    async def cfg_reset(self):
        config = copy.deepcopy(self._config)
        if os.path.isfile(PROVIDER_CONFIG):
            try:
                os.remove(PROVIDER_CONFIG)
            except:
                pass
        # removing objects from config:
        config['instances']    = []
        config['environments'] = []
        config['jobs']         = []

        self._init(config)        

    @abstractmethod
    async def deploy(self):
        pass

    @abstractmethod
    async def run(self):
        pass

    @abstractmethod
    async def wait(self,job_state,run_session=None):
        pass

    @abstractmethod
    async def get_jobs_states(self,run_session=None):
        pass

    @abstractmethod
    async def print_jobs_summary(self,run_session=None,instance=None):
        pass

    @abstractmethod
    async def print_aborted_logs(self,run_session=None,instance=None):
        pass

    @abstractmethod
    async def print_objects(self):
        pass

    @abstractmethod
    async def fetch_results(self,directory,run_session=None):
        pass
          
    @abstractmethod
    def get_suggested_image(self,region):
        pass

def get_config(path):
    config_file = path
    configdir   = os.path.dirname(config_file)
    configbase  = os.path.basename(config_file)
    config_name , config_extension = os.path.splitext(configbase)

    if config_extension=='.json':
        try:
            if os.path.exists(config_file):
                with open(config_file,'r') as config_file:
                    config = json.loads(config_file.read())
                print("loaded config from json file")
            else:
                config = None
        except:
            config = None
    elif config_extension=='.py':
        try:
            sys.path.append(os.path.abspath(configdir))
            configModule = __import__(config_name,globals(),locals())
            config = configModule.config
        except ModuleNotFoundError as mfe:
            config = None
        except:
            config = None
    return config

def get_client(config_=None):

    if isinstance(config_,str): 
        config = get_config(config_)
    elif isinstance(config_,dict):
        config = config_
    else: # get the auto-saved config
        config = get_config(PROVIDER_CONFIG)

    if config is None:
        debug(1,"You need to specify a valid configuration path",color=bcolors.FAIL)
        debug(1,"(could also not find default config file",PROVIDER_CONFIG,")",color=bcolors.FAIL)
        raise CloudSendError()

    if config.get('provider','aws') == 'aws':

        if config.get('maestro','local') == 'local':
            
            craws  = __import__("cloudsend.aws")

            client = craws.aws.AWSCloudSendFatProvider(config)

            return client

        else:

            craws  = __import__("cloudsend.aws")

            client = craws.aws.AWSCloudSendLightProvider(config)

            return client

    else:

        debug(1,config.get('provider'), " not implemented yet")

        raise CloudSendError()

def line_buffered(f):
    line_buf = ""
    doContinue = True
    try :
        while doContinue and not f.channel.get_exit_signal():
            try:
                line_buf += f.read(1).decode("utf-8")
                if line_buf.endswith('\n'):
                    yield line_buf
                    line_buf = ''
            except Exception as e:
                #errmsg = str(e)
                #debug(1,"error (1) while buffering line",errmsg)
                pass
                #doContinue = False
    except Exception as e0:
        debug(1,"error (2) while buffering line",str(e0))
        #doContinue = False

def escape_arg_for_send(args):
    if not args:
        return args
    args_escaped = []
    for i,arg in enumerate(args):
        if isinstance(arg,dict):
            arg = json.dump(arg)
        elif not isinstance(arg,str):
            arg = str(arg)
        arg = arg.replace("\"","\\\"")
        args_escaped.append(arg)
    return args_escaped  

def make_client_command(maestro_command,args):
    if not args:
        the_command = maestro_command
    else:
        if not isinstance(args,list):
            args = [ args ]
        args = escape_arg_for_send(args)
        the_command = COMMAND_ARGS_SEP.join( [ maestro_command , ARGS_SEP.join(args) ] )
    return the_command

DBG_LVL=1
DBG_PREFIX=None

def debug(level,*args,**kwargs):
    if level <= DBG_LVL:
        if 'color' in kwargs:
            color = kwargs['color']
            listargs = list(args)
            #listargs.insert(0,color)
            if isinstance(listargs[0], (bytes, bytearray)):
                str_concat = listargs[0].decode()
            else:
                str_concat = str(listargs[0])
            listargs[0] = color + str_concat
            listargs.append(bcolors.ENDC)
            args = tuple(listargs)
            kwargs.pop('color')
        try:
            if not sys.stdout: # or sys.stdout.closed==True:
                return

            if DBG_PREFIX:
                print(DBG_PREFIX,*args,**kwargs)
            else:
                print(*args,**kwargs)
            sys.stdout.flush()
            sys.stderr.flush()
        except:
            pass



from abc import ABC , abstractmethod
from cloudsend.provider import CloudSendProvider , line_buffered , make_client_command
from cloudsend.core import *
import copy , io
from zipfile import ZipFile
import os , fnmatch
import re , json
from os.path import basename
import pkg_resources
import time
import random
import shutil

random.seed()

####################################
# Client handling MAESTRO instance #
####################################

class CloudSendLightProvider(CloudSendProvider,ABC):

    def __init__(self,conf=None):

        CloudSendProvider.__init__(self,conf)

        #self._install_maestro()

    def _init(self,conf):
        self.ssh_conn   = None
        self.ftp_client = None
        super()._init(conf)
        self._load()
    
    def _load(self):
        self._maestro = None
        if self._config.get('maestro')=='remote':
            if not self._config.get('instances') or len(self._config.get('instances'))==0:
                self.debug(2,"There are no instances to watch - skipping maestro creation")
                return
            else:
                # let's try to match the region of the first instance ...
                region = None
                if len(self._config.get('instances')) > 0:
                    region = self._config.get('instances')[0].get('region')
                if not region:
                    region = self.get_region()

                img_id , img_user , img_type = self.get_suggested_image(region)
                if not img_id:
                    self.debug(1,"Using first instance information to create MAESTRO")
                    img_id   = self._config.get('instances')[0].get('img_id')
                    img_user = self._config.get('instances')[0].get('img_username')
                    img_type = self._config.get('instances')[0].get('img_type')
                
                maestro_cfg = { 
                    'maestro'      : True ,
                    'img_id'       : img_id ,
                    'img_username' : img_user ,                 
                    'type'         : img_type , 
                    'dev'          : self._config.get('dev',False) ,
                    'project'      : self._config.get('project',None) ,
                    'region'       : region
                }
                self._maestro = CloudSendInstance(maestro_cfg,None)

    async def _deploy_maestro(self,reset):
        # deploy the maestro ...
        if self._maestro is None:
            self.debug(2,"no MAESTRO object to deploy")
            return

        # wait for maestro to be ready
        instanceid , ssh_conn , ftp_client = await self._wait_and_connect(self._maestro)

        home_dir = self._maestro.get_home_dir()
        cloudsend_dir = self._maestro.path_join( home_dir , 'cloudsend' )
        files_dir = self._maestro.path_join( cloudsend_dir , 'files' )
        ready_file = self._maestro.path_join( cloudsend_dir , 'ready' )
        maestro_file = self._maestro.path_join( home_dir , 'maestro' )
        aws_dir = self._maestro.path_join( home_dir , '.aws' )
        aws_config = self._maestro.path_join( aws_dir , 'config' )
        if self._maestro.get_platform() == CloudSendPlatform.LINUX or self._maestro.get_platform() == CloudSendPlatform.WINDOWS_WSL :
            activate_file = self._maestro.path_join( cloudsend_dir , '.venv' , 'maestro' , 'bin' , 'activate' )
        elif self._maestro.get_platform() == CloudSendPlatform.WINDOWS:
            activate_file = self._maestro.path_join( cloudsend_dir , '.venv' , 'maestro' , 'Scripts' , 'activate.bat' )
        
        re_init  = await self._test_reupload(self._maestro,ready_file, ssh_conn)

        if re_init:
            # remove the file
            await self._exec_command(ssh_conn,'rm -f ' + ready_file )
            # make cloudsend dir
            await self._exec_command(ssh_conn,'mkdir -p ' + files_dir ) 
            # mark it as maestro...
            await self._exec_command(ssh_conn,'echo "" > ' + maestro_file ) 
            # add manually the 
            if self._config.get('profile'):
                profile = self._config.get('profile')
                region  = self.get_region() # we have a profile so this returns the region for this profile
                aws_config_cmd = "mkdir -p "+aws_dir+" && echo \"[profile "+profile+"]\nregion = " + region + "\noutput = json\" > " + aws_config
                await self._exec_command(ssh_conn,aws_config_cmd)
            # grant its admin rights (we need to be (stopped or) running to be able to do that)
            self.grant_admin_rights(self._maestro)
            # setup auto_stop behavior for maestro
            self.setup_auto_stop(self._maestro)
            # deploy CloudSend on the maestro
            await self._deploy_cloudsend(ssh_conn,ftp_client)
            # mark as ready
            await self._exec_command(ssh_conn,'if [ -f '+activate_file+' ]; then echo "" > '+ready_file+' ; fi')

        # deploy the config to the maestro (every time)
        await self._deploy_config(ssh_conn,ftp_client)

        # let's redeploy the code every time for now ... (if not already done in re_init)
        if not re_init and reset:
            await self._deploy_cloudsend_files(ssh_conn,ftp_client)

        # start the server (if not already started)
        await self._run_server(ssh_conn)

        # wait for maestro to have started
        await self._wait_for_maestro(ssh_conn)

        #time.sleep(30)

        self.ssh_conn = ssh_conn
        self.ftp_client = ftp_client
        #ftp_client.close()
        #ssh_conn.close()

        self.debug(1,"MAESTRO is READY",color=bcolors.OKCYAN)


    async def _deploy_cloudsend_files(self,ssh_conn,ftp_client):
        await ftp_client.chdir(self._get_cloudsend_dir())
        # CP437 is IBM zip file encoding
        await self.sftp_put_bytes(ftp_client,'cloudsend.zip',self._create_cloudsend_zip(),'CP437')
        commands = [
            { 'cmd' : 'cd '+self._get_cloudsend_dir()+' && unzip -o cloudsend.zip && rm cloudsend.zip' , 'out' : True } ,
        ]
        await self._run_ssh_commands(self._maestro,ssh_conn,commands) 
        
        filesOfDirectory = os.listdir('.')
        pattern = "cloudsend*.pem"
        for file in filesOfDirectory:
            if fnmatch.fnmatch(file, pattern):
                await ftp_client.put(os.path.abspath(file),os.path.basename(file))
        sh_files = self._get_remote_files_path( '*.sh')
        commands = [
            { 'cmd' : 'chmod +x ' + sh_files , 'out' : True } ,
        ]
        await self._run_ssh_commands(self._maestro,ssh_conn,commands) 


    async def _deploy_cloudsend(self,ssh_conn,ftp_client):

        # upload cloudsend files
        await self._deploy_cloudsend_files(ssh_conn,ftp_client)
        
        maestroenv_sh = self._get_remote_files_path( 'maestroenv.sh' )
        # unzip cloudsend files, install and run
        commands = [
            { 'cmd' : maestroenv_sh , 'out' : True } ,
        ]
        await self._run_ssh_commands(self._maestro,ssh_conn,commands)

    async def _run_server(self,ssh_conn):
        # run the server
        startmaestro_sh = self._get_remote_files_path( 'startmaestro.sh' )
        commands = [
            { 'cmd' : startmaestro_sh , 'out' : False , 'output' : 'maestro.log' },
            { 'cmd' : 'crontab -r ; echo "* * * * * '+startmaestro_sh+' auto_init" | crontab', 'out' : True }
        ]
        await self._run_ssh_commands(self._maestro,ssh_conn,commands)

        # we have separated run and init and start 
        # we need to init with the deployed config
        self.debug(1,"initialize server with config")
        await self._exec_maestro_command("init","config.json")

    async def _deploy_config(self,ssh_conn,ftp_client,config_filename='config.json'):
        config , mkdir_cmd , files_to_upload_per_dir = self._translate_config_for_maestro()
        # serialize the config and send it to the maestro
        await ftp_client.chdir(self._get_cloudsend_dir())
        await self.sftp_put_string(ftp_client,config_filename,json.dumps(config))
        # execute the mkdir_cmd
        if mkdir_cmd:
            await self._exec_command(ssh_conn,mkdir_cmd)
        for remote_dir , files_infos in files_to_upload_per_dir.items():
            await ftp_client.chdir(remote_dir)
            for file_info in files_infos:
                remote_file = os.path.basename(file_info['remote'])
                try:
                    await ftp_client.put(file_info['local'],remote_file)
                except FileNotFoundError as fnfe:
                    self.debug(1,"You have specified a file that does not exist:",fnfe,color=bcolors.FAIL)
        return config 

    def _translate_config_for_maestro(self):
        config = copy.deepcopy(self._config)

        config['maestro'] = 'local' # the config is for maestro, which needs to run local

        config['region'] = self.get_region()

        for i , inst_cfg in enumerate(config.get('instances')):
            # we need to freeze the region within each instance config ...
            if not config['instances'][i].get('region'):
                config['instances'][i]['region'] = self.get_region()

        for i , env_cfg in enumerate(config.get('environments')):
            # Note: we are using a simple CloudSendEnvironment here 
            # we're not deploying at this stage
            # but this will help use re-use all the business logic in cloudsendutils envs
            # in order to standardize the environment and create an inline version 
            # through the CloudSendEnvironment::json() method
            env = CloudSendEnvironment(self._config.get('project'),env_cfg)
            config['environments'][i] = env.get_env_obj()
            config['environments'][i][K_COMPUTED] = True

        files_to_upload = dict()

        for i , job_cfg in enumerate(config.get('jobs')):
            job = CloudSendJob(job_cfg,i)
            run_script = job.get_config('run_script')  
            ref_file0  = run_script
            args       = ref_file0.split(' ')
            if args:
                ref_file0 = args[0]
            #for attr,use_ref in [ ('run_script',False) , ('upload_files',True) , ('input_file',True) , ('output_file',True) ] :
            for attr,use_ref in [ ('run_script',False) , ('upload_files',True) , ('input_file',True) ] :
                upfiles = job_cfg.get(attr)
                ref_file = ref_file0 if use_ref else None
                if not upfiles:
                    continue
                if isinstance(upfiles,str):
                    upfiles_split = upfiles.split(' ')
                    if len(upfiles_split)>0:
                        upfiles = upfiles_split[0]
                if isinstance(upfiles,str):
                    local_abs_path , local_rel_path , remote_abs_path , remote_rel_path , external = self._resolve_maestro_job_paths(upfiles,ref_file,self._get_home_dir())
                    if local_abs_path not in files_to_upload:
                        files_to_upload[local_abs_path] = { 'local' : local_abs_path , 'remote' : remote_abs_path }
                    if attr == 'run_script':
                        args = run_script.split(' ')
                        args[0] = remote_abs_path
                        config['jobs'][i][attr] = (' ').join(args)
                    else:
                        config['jobs'][i][attr] = remote_abs_path
                else:
                    config['jobs'][i][attr] = []
                    for upfile in upfiles:
                        local_abs_path , local_rel_path , remote_abs_path , remote_rel_path , external = self._resolve_maestro_job_paths(upfile,ref_file,self._get_home_dir())
                        if local_abs_path not in files_to_upload:
                            files_to_upload[local_abs_path] = { 'local' : local_abs_path , 'remote' : remote_abs_path }
                        config['jobs'][i][attr].append(remote_abs_path)
        
        mkdir_cmd = ""
        files_to_upload_per_dir = dict()
        for key,file_info in files_to_upload.items():
            remote_dir = os.path.dirname(file_info['remote'])
            if remote_dir not in files_to_upload_per_dir:
                files_to_upload_per_dir[remote_dir] = []
                mkdir_cmd += "mkdir -p " + remote_dir + ";"
            files_to_upload_per_dir[remote_dir].append(file_info)

        self.debug(2,config)

        return config , mkdir_cmd , files_to_upload_per_dir

    def _resolve_maestro_job_paths(self,upfile,ref_file,home_dir):
        files_path = self._maestro.path_join( home_dir , 'files' )
        return cloudsendutils.resolve_paths(self._maestro,upfile,ref_file,files_path,True) # True for mutualized 

    def _get_home_dir(self):
        return self._maestro.get_home_dir()

    def _get_cloudsend_dir(self):
        return self._maestro.path_join( self._get_home_dir() , 'cloudsend' )

    def _get_remote_files_path(self,files=None):
        if not files:
            return self._maestro.path_join( self._get_cloudsend_dir() , 'cloudsend' , 'resources' , 'remote_files' )        
        else:
            return self._maestro.path_join( self._get_cloudsend_dir() , 'cloudsend' , 'resources' , 'remote_files' , files )        

    def _zip_package(self,package_name,src,dest,zipObj):
        dest = os.path.join( dest , os.path.basename(src) ).rstrip( os.sep )
        if pkg_resources.resource_isdir(package_name, src):
            #if not os.path.isdir(dest):
            #    os.makedirs(dest)
            for res in pkg_resources.resource_listdir(package_name, src):
                self.debug(2,'scanning package resource',res)
                self._zip_package(package_name,os.path.join( src , res), dest, zipObj)
        else:
            if os.path.splitext(src)[1] not in [".pyc"] and not src.strip().endswith(".DS_Store"):
                #copy_resource_file(src, dest) 
                data_str = pkg_resources.resource_string(package_name, src)
                self.debug(2,"Writing",src)
                zipObj.writestr(dest,data_str)

    def _create_cloudsend_zip(self):
        zip_buffer = io.BytesIO()
        # create a ZipFile object
        with ZipFile(zip_buffer, 'w') as zipObj:    
            # for otherfilepath in ['pyproject.toml' , 'requirements.txt' ]:
            #     with open(otherfilepath,'r') as thefile:
            #         data = thefile.read()
            #         zipObj.writestr(otherfilepath,data)
            # write the package
            self._zip_package("cloudsend",".","cloudsend",zipObj)

        self.debug(2,"ZIP BUFFER SIZE =",zip_buffer.getbuffer().nbytes)

        # very important...
        zip_buffer.seek(0)

        return zip_buffer.getvalue() #.getbuffer()

    async def _wait_for_maestro(self,ssh_conn):
        #if self.ssh_conn is None:
        #    instanceid , self.ssh_conn , self.ftp_client = self._wait_and_connect(self._maestro)
        
        waitmaestro_sh = self._get_remote_files_path( 'waitmaestro.sh' )
        cmd = waitmaestro_sh+" 1" # 1 = with tail log

        stdout , stderr = await self._exec_command(ssh_conn,cmd)      

        try:
            async for line in stdout:
                if not line:
                    break
                self.debug(1,line,end='')
        except StopAsyncOperation:
            pass

        # for l in await line_buffered(stdout):
        #     if not l:
        #         break
        #     self.debug(1,l,end='')

    async def _exec_maestro_command(self,maestro_command,args=None):

        the_command = make_client_command(maestro_command,args)

        if self.ssh_conn is None:
            instanceid , self.ssh_conn , self.ftp_client = await self._wait_and_connect(self._maestro)
        
        private_ip = self._maestro.get_ip_addr_priv()

        # -u for no buffering
        waitmaestro_sh = self._get_remote_files_path( 'waitmaestro.sh' )
        venv_python    = self._maestro.path_join( self._get_cloudsend_dir() , '.venv' , 'maestro' , 'bin' , 'python3' )

        cmd = "cd "+self._get_cloudsend_dir()+ " && "+waitmaestro_sh+" && sudo "+venv_python+" -u -m cloudsend.maestroclient " + the_command

        stdout , stderr = await self._exec_command(self.ssh_conn,cmd)

        try:
            async for line in stdout:
                if not line:
                    break
                self.debug(1,line,end='')
        except StopAsyncOperation:
            pass
        except:
            pass

        # for l in await line_buffered(stdout):
        #     if not l:
        #         break
        #     self.debug(1,l,end='')
        
        # while True:
        #     outlines = stdout.readlines()
        #     if not outlines:
        #         errlines = stderr.readlines()
        #         for eline in errlines:
        #             self.debug(1,eline,end='')
        #         break
        #     for line in outlines:
        #         self.debug(1,line,end='')
        #     errlines = stderr.readlines()
        #     for eline in errlines:
        #         self.debug(1,eline,end='')

        # for l in line_buffered(stdout):
        #     if not l:
        #         errlines = stderr.readlines()
        #         for eline in errlines:
        #             self.debug(1,eline,end='')
        #         break
        #     self.debug(1,l,end='')
        #     errlines = stderr.readlines()
        #     for eline in errlines:
        #         self.debug(1,eline,end='')


    async def _install_maestro(self,reset):
        # this will block for some time the first time...
        self.debug_set_prefix(bcolors.BOLD+'INSTALLING MAESTRO: '+bcolors.ENDC)
        self._instances_states = dict()        
        self._start_and_update_instance(self._maestro)
        if reset:
            await self.reset_instance(self._maestro)
        await self._deploy_maestro(reset) # deploy the maestro now !
        self.debug_set_prefix(None)       

    async def start(self,reset=False):
        # install maestro materials
        await self._install_maestro(reset)
        # triggers maestro::start
        await self._exec_maestro_command("start",reset)

    async def _cfg_add_objects(self,conf,coroutine,name):
        # complement self._config
        await coroutine(conf)
        # wait for maestro to be ready
        instanceid , ssh_conn , ftp_client = await self._wait_and_connect(self._maestro)
        # config_name
        config_name = 'config_add-' + cloudsendutils.generate_unique_id() + '.json'
        # deploy the new config to the maestro (every time) (including dependent files)
        translated_config = await self._deploy_config(ssh_conn,ftp_client,config_name)
        # triggers maestro::add_* : use string dump or config file name (either way)
        #await self._exec_maestro_command(name,config_name)
        await self._exec_maestro_command(name,json.dumps(translated_config))

    async def cfg_add_instances(self,conf):
        await self._cfg_add_objects(conf,super().cfg_add_instances,"cfg_add_instances")

    async def cfg_add_environments(self,conf):
        await self._cfg_add_objects(conf,super().cfg_add_environments,"cfg_add_environments")

    async def cfg_add_jobs(self,conf):
        await self._cfg_add_objects(conf,super().cfg_add_jobs,"cfg_add_jobs")

    async def cfg_add_config(self,conf):
        await self._cfg_add_objects(conf,super().cfg_add_config,"cfg_add_config")

    async def cfg_reset(self):
        # triggers maestro::reset
        await self._exec_maestro_command("cfg_reset") # use output - the deploy part will be skipped depending on option ...

    async def reset_instance(self,instance):
        self.debug(1,'RESETTING instance',instance.get_name())
        instanceid, ssh_conn , ftp_client = await self._wait_and_connect(instance)
        if ssh_conn is not None:
            await self.sftp_put_remote_file(ftp_client,'resetmaestro.sh')
            resetmaestro_sh = self._maestro.path_join( self._maestro.get_home_dir() , 'resetmaestro.sh' )
            commands = [
               { 'cmd' : 'chmod +x '+resetmaestro_sh+' && ' + resetmaestro_sh , 'out' : True }
            ]
            await self._run_ssh_commands(instance,ssh_conn,commands)
            #ftp_client.close()
            ssh_conn.close()
        self.debug(1,'RESETTING done')

    async def deploy(self):
        # triggers maestro::deploy
        await self._exec_maestro_command("deploy") # use output - the deploy part will be skipped depending on option ...

    async def run(self):
        # triggers maestro::run
        await self._exec_maestro_command("run")

    async def wakeup(self):
        await self.start()
        # triggers maestro::wakeup
        await self._exec_maestro_command("wakeup")

    async def wait(self,job_state,run_session=None):
        # triggers maestro::wait
        await self._exec_maestro_command("wait")

    async def get_jobs_states(self,run_session=None):
        # triggers maestro::get_jobs_states
        await self._exec_maestro_command("get_states")

    async def print_jobs_summary(self,run_session=None,instance=None):
        # triggers maestro::print_jobs_summary
        await self._exec_maestro_command("print_summary")

    async def print_aborted_logs(self,run_session=None,instance=None):
        # triggers maestro::print_aborted_logs
        await self._exec_maestro_command("print_aborted")

    async def print_objects(self):
        # triggers maestro::print_objects
        await self._exec_maestro_command("print_objects")

    async def fetch_results(self,out_dir,run_session=None):

        try:
            #os.rmdir(out_dir)
            shutil.rmtree(out_dir, ignore_errors=True)
        except:
            pass
        try:
            os.makedirs(out_dir)
        except:
            pass

        randnum          = str(random.randrange(1000))
        homedir          = self._maestro.get_home_dir()
        maestro_dir      = self._maestro.path_join( homedir , "cloudsend_tmp_fetch" + randnum )
        maestro_tar_file = "maestro"+randnum+".tar"
        maestro_tar_path = self._maestro.path_join( homedir , maestro_tar_file )

        # fetch the results on the maestro
        await self._exec_maestro_command("fetch_results",maestro_dir)

        # get the tar file of the results
        instanceid , ssh_conn , ftp_client = await self._wait_and_connect(self._maestro)
        stdout,stderr = await self._exec_command(ssh_conn,"cd " + maestro_dir + " && tar -cvf "+maestro_tar_path+" .")      
        stdout_str = await stdout.read()
        stderr_str = await stderr.read()
        self.debug(1,stdout_str) 
        self.debug(2,stderr_str) 
        local_tar_path = os.path.join(out_dir,maestro_tar_file)
        await ftp_client.chdir( homedir )
        await ftp_client.get( maestro_tar_file , local_tar_path )

        # untar
        os.system("tar -xvf "+local_tar_path+" -C "+out_dir)

        # cleanup
        os.remove(local_tar_path)
        stdout,stderr = await self._exec_command(ssh_conn,"rm -rf "+maestro_dir+" "+maestro_tar_path)

        # close
        #ftp_client.close()
        ssh_conn.close()

    def _get_or_create_instance(self,instance):
        instance , created = super()._get_or_create_instance(instance)
        # dangerous !
        #self.add_maestro_security_group(instance)
        # open web server

        return instance , created 

    @abstractmethod
    def grant_admin_rights(self,instance):
        pass

    @abstractmethod
    def add_maestro_security_group(self,instance):
        pass

    @abstractmethod
    def setup_auto_stop(self,instance):
        pass

    # needed by CloudSendProvider::_wait_for_instance 
    def serialize_state(self):
        pass






import copy
import sys
import os , signal
import json
import pickle
import math
from cloudsend.core import *
from datetime import date, datetime
import traceback

class ConfigManager():

    def __init__(self,provider,conf,instances,environments,jobs):
        
        self._provider = provider
        self._config = conf
        self._instances = instances
        self._environments = environments
        self._jobs = jobs

    def load(self):
        self._load_objects()
        self._preprocess_jobs()
        self._sanity_checks()

    def _load_objects(self):
        projectName = self._config.get('project')
        inst_cfgs   = self._config.get('instances')
        env_cfgs    = self._config.get('environments')
        job_cfgs    = self._config.get('jobs')

        #self._instances = [ ]
        if inst_cfgs:
            for inst_cfg in inst_cfgs:
                if inst_cfg.get(K_LOADED) == True:
                    continue 

                # virtually demultiply according to 'number' and 'explode'
                for i in range(inst_cfg.get('number',1)):
                    rec_cpus = self._provider.get_recommended_cpus(inst_cfg)
                    if rec_cpus is None:
                        self._provider.debug(1,"WARNING: could not set recommended CPU size for instance type:",inst_cfg.get('type'))
                        cpu_split = None
                    else:
                        cpu_split = rec_cpus[len(rec_cpus)-1]

                    if cpu_split is None:
                        if inst_cfg.get('cpus') is not None:
                            oldcpu = inst_cfg.get('cpus')
                            self._provider.debug(1,"WARNING: removing CPU reqs for instance type:",inst_cfg.get('type'),"| from",oldcpu,">>> None")
                        real_inst_cfg = copy.deepcopy(inst_cfg)
                        real_inst_cfg.pop('number',None)  # provide default value to avoid KeyError
                        real_inst_cfg.pop('explode',None) # provide default value to avoid KeyError

                        real_inst_cfg['cpus'] = None
                        real_inst_cfg['rank'] = "{0}.{1}".format(i+1,1)

                        self._load_other_instance_info(real_inst_cfg)

                        # starts calling the Service here !
                        # instance , created = self.start_instance( real_inst_cfg )
                        # let's put some dummy instances for now ...
                        instance = CloudSendInstance( real_inst_cfg , None, None )
                        self._instances.append( instance )
                    else:
                        total_inst_cpus = inst_cfg.get('cpus',1)
                        if type(total_inst_cpus) != int and type(total_inst_cpus) != float:
                            cpucore = self._provider.get_cpus_cores(inst_cfg)
                            total_inst_cpus = cpucore if cpucore is not None else 1
                            self._provider.debug(1,"WARNING: setting default CPUs number to",cpucore,"for instance",inst_cfg.get('type'))
                        if not inst_cfg.get('explode') and total_inst_cpus > cpu_split:
                            self._provider.debug(1,"WARNING: forcing 'explode' to True because required number of CPUs",total_inst_cpus,' is superior to ',inst_cfg.get('type'),'max number of CPUs',cpu_split)
                            inst_cfg.set('explode',True)

                        if inst_cfg.get('explode'):
                            num_sub_instances = math.floor( total_inst_cpus / cpu_split ) # we target CPUs with 16 cores...
                            if num_sub_instances == 0:
                                cpu_inc = total_inst_cpus
                                num_sub_instances = 1
                            else:
                                cpu_inc = cpu_split 
                                num_sub_instances = num_sub_instances + 1
                        else:
                            num_sub_instances = 1
                            cpu_inc = total_inst_cpus
                        cpus_created = 0 
                        for j in range(num_sub_instances):
                            rank = "{0}.{1}".format(i+1,j+1)
                            if j == num_sub_instances-1: # for the last one we're completing the cpus with whatever
                                inst_cpus = total_inst_cpus - cpus_created
                            else:
                                inst_cpus = cpu_inc
                            if inst_cpus == 0:
                                continue 

                            if rec_cpus is not None and not inst_cpus in rec_cpus:
                                self._provider.debug(1,"ERROR: The total number of CPUs required causes a sub-number of CPUs ( =",inst_cpus,") to not be accepted by the type",inst_cfg.get('type'),"| list of valid cpus:",rec_cpus)
                                sys.exit()

                            real_inst_cfg = copy.deepcopy(inst_cfg)
                            real_inst_cfg.pop('number',None)  # provide default value to avoid KeyError
                            real_inst_cfg.pop('explode',None) # provide default value to avoid KeyError
                            real_inst_cfg['cpus'] = inst_cpus
                            real_inst_cfg['rank'] = rank

                            self._load_other_instance_info(real_inst_cfg)

                            # let's put some dummy instances for now ...
                            instance = CloudSendInstance( real_inst_cfg , None, None )
                            self._instances.append( instance )

                            cpus_created = cpus_created + inst_cpus

                    inst_cfg[K_LOADED] = True
        
        self._provider.debug(3,self._instances)

        #self._environments = [ ] 
        if env_cfgs:
            for env_cfg in env_cfgs:
                if env_cfg.get(K_LOADED) == True:
                    continue

                # copy the dev global paramter to the environment configuration (will be used for names)
                env_cfg['dev']  = self._config.get('dev',False)
                env = CloudSendEnvironment(projectName,env_cfg)
                self._environments.append(env)

                env_cfg[K_LOADED] = True

        #self._jobs = [ ] 
        if job_cfgs:
            rank=0
            for i,job_cfg in enumerate(job_cfgs):
                if job_cfg.get(K_LOADED) == True:
                    continue 

                for j in range(job_cfg.get('repeat',1)):
                    job = CloudSendJob(job_cfg,rank)
                    self._jobs.append(job)
                    rank = rank + 1
                
                job_cfg[K_LOADED] = True

    def _load_other_instance_info(self,real_inst_cfg):
        # [!important!] also copy global information that are used for the name generation ...
        real_inst_cfg['dev']      = self._config.get('dev',False)
        real_inst_cfg['project']  = self._config.get('project',None)

        # let's also freeze the region
        # this is done so that we dont change the hash when sending the config to maestro
        # (because the client adds the 'region' field when translating the config)
        if not real_inst_cfg.get('region'):
            real_inst_cfg['region'] = self._provider.get_region()     

        # also set the img_id is None
        if not real_inst_cfg.get('img_id'):
            img_id , img_username , img_type = self._provider.get_suggested_image(real_inst_cfg['region'])
            real_inst_cfg['img_id']          = img_id 
            real_inst_cfg['img_username']    = img_username 

    # fill up the jobs names if not present (and we have only 1 environment defined)
    # link the jobs objects with an environment object
    def _preprocess_jobs(self):
        for job in self._jobs:
            if not job.get_config('env_name'):
                if len(self._environments)==1:
                    job.attach_env(self._environments[0])
                else:
                    print("FATAL ERROR - you have more than one environments defined and the job doesnt have an env_name defined",job)
                    sys.exit()
            else:
                env = self._get_environment(job.get_config('env_name'))
                if not env:
                    print("FATAL ERROR - could not find env with name",job.get_config('env_name'),job)
                    sys.exit()
                else:
                    job.attach_env(env)

    def _sanity_checks(self):
        if len(self._instances) > 10:
            self._provider.debug(1,"\033[91mWATCH OUT ! You are creating more than 10 instances - not allowed for now!\033[0m")
            raise Exception()

    def _get_environment(self,name):
        for env in self._environments:
            if env.get_name() == name:
                return env
        return None

# def json_get_key(obj):
#     if isinstance(obj,CloudSendInstance):
#         return "__cloudsend_instance:" + obj.get_name()
#     elif isinstance(obj,CloudSendDeployedEnvironment):
#         return "__cloudsend_environment_dpl:" + obj.get_name()
#     elif isinstance(obj,CloudSendEnvironment):
#         return "__cloudsend_environment:" + obj.get_name()
#     elif isinstance(obj,CloudSendDeployedJob):
#         return "__cloudsend_job_dpl:" + obj.get_hash() + "|" + str(obj.get_job().get_rank()) + "," + str(obj.get_rank())
#     elif isinstance(obj,CloudSendJob):
#         return "__cloudsend_job:" + obj.get_hash() + "|" + str(obj.get_rank()) 
#     elif isinstance(obj,CloudSendProcess):
#         return "__cloudsend_process:" + obj.get_uid()
#     else:
#         return str(cs_obj)

# class CloudSendJSONEncoder(json.JSONEncoder):

#     def __init__(self, *args, **kwargs):
#         kwargs['check_circular'] = False  # no need to check anymore
#         super(CloudSendJSONEncoder,self).__init__(*args, **kwargs)
#         self.proc_objs = []

#     def default(self, obj):

#         if  isinstance(obj, (CloudSendInstance, CloudSendEnvironment, CloudSendDeployedEnvironment, CloudSendJob, CloudSendDeployedJob, CloudSendProcess)):
#             if obj in self.proc_objs:
#                 return json_get_key(obj)
#             else:
#                 self.proc_objs.append(obj)
#             return { **obj.__dict__ , **{'__class__':type(obj).__name__} }
        
#         elif isinstance(obj, (datetime, date)):
#             return obj.isoformat()  # Let the base class default method raise the TypeError

#         return super(CloudSendJSONEncoder,self).default(obj) #json.JSONEncoder.default(self, obj)

# class CloudSendJSONDecoder(json.JSONDecoder):
#     def __init__(self, *args, **kwargs):
#         json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)
#         self._references = dict()
    
#     def object_hook(self, dct):
#         #print(dct)
#         if '__class__' in dct:
#             class_name = dct['__class__']
#             try :
#                 if class_name == 'CloudSendInstance':
#                     obj = CloudSendInstance(dct['_config'],dct['_id'],dct['_data'])
#                     obj.__dict__.update(dct)
#                 elif class_name == 'CloudSendEnvironment':
#                     obj = CloudSendEnvironment(dct['_project'],dct['_config'])
#                     obj.__dict__.update(dct)
#                 elif class_name == 'CloudSendDeployedEnvironment':
#                     env_ref = None #self._references[dct['_env']]
#                     ins_ref = None #self._references[dct['_instance']]
#                     obj = CloudSendDeployedEnvironment.__new__(CloudSendDeployedEnvironment) #CloudSendDeployedEnvironment(env_ref,ins_ref)
#                     obj.__dict__.update(dct)
#                 elif class_name == 'CloudSendJob':
#                     obj = CloudSendJob(dct['_config'],dct['_rank'])
#                     obj.__dict__.update(dct)
#                 elif class_name == 'CloudSendDeployedJob':
#                     job_ref = None #self._references[dct['_job']]
#                     env_ref = None #self._references[dct['_env']]
#                     obj = CloudSendDeployedJob.__new__(CloudSendDeployedJob) #CloudSendDeployedJob(job_ref,env_ref)
#                     obj.__dict__.update(dct)
#                 elif class_name == 'CloudSendProcess':
#                     job_ref = None #self._references[dct['_job']]
#                     obj = CloudSendProcess.__new__(CloudSendProcess) #CloudSendProcess(job_ref,dct['_uid'],dct['_pid'],dct['_batch_uid'])
#                     obj.__dict__.update(dct)
#                 self._references[json_get_key(obj)] = obj
#                 return obj
#             except KeyError as ke:
#                 traceback.print_exc()
#                 print("KEY ERROR while DESERIALIZATION",ke)
#                 print(self._references)
#                 return None
#         else:
#             return dct        


STATE_FILE = 'state.pickle'

class StateSerializer():

    def __init__(self,provider):
        self._provider = provider

        self._state_file = STATE_FILE
        self._loaded = None

    def reset(self):
        if os.path.isfile(self._state_file):
            try:
                os.remove(self._state_file)
            except:
                pass

    def serialize(self,provider_state,instances,environments,jobs,run_sessions,current_session):
        try:
            state = {
                'state' : provider_state ,
                'instances' : instances ,
                'environments' : environments ,
                'jobs' : jobs ,
                'run_sessions' : run_sessions ,
                'current_session' : current_session
            }
            for job in state['jobs']:
                lastp = job.get_last_process()
                self._provider.debug(3,"PROCESS TO SERIALIZE",lastp)

            #json_data = json.dumps(state,indent=4,cls=CloudSendJSONEncoder)
            with open(self._state_file,'wb') as state_file:
                pickle.dump(state,state_file)#,protocol=0) # protocol 0 = readable
                #state_file.write(json_data)
        except Exception as e:
            self._provider.debug(1,"SERIALIZATION ERROR",e)


    def load(self):
        if not os.path.isfile(self._state_file):
            self._provider.debug(2,"StateSerializer: no state serialized")
            return False
        try:
            with open(self._state_file,'rb') as state_file:
                #json_data = state_file.read()
                #objects   = json.loads(json_data,cls=CloudSendJSONDecoder)
                self._loaded = pickle.load(state_file)
        except Exception as e:
            traceback.print_exc()
            self._provider.debug(1,"DE-SERIALIZATION ERROR",e)

    # check if the state is consistent with the provider objects that have been loaded from the config...
    # TODO: do that
    def check_consistency(self,state,instances,environments,jobs,run_sessions,current_session):
        if self._loaded is None:
            self._provider.debug(2,"Seralized data not loaded. No consistency")
            return False 
        try:
            _state        = self._loaded['state']
            _instances    = self._loaded['instances']
            _environments = self._loaded['environments']
            _jobs         = self._loaded['jobs']
            # those won't be compared because they are run time memory stuff...
            _run_sessions = self._loaded['run_sessions']
            _current_session = self._loaded['current_session']
            for job in jobs:
                lastp = job.get_last_process()
                self._provider.debug(3,"DESERIALIZED PROCESS",lastp)
            # state wont be the same by definition ...
            #assert state==_state 
            assert len(instances)==len(_instances)
            assert len(environments)==len(_environments)
            assert len(jobs)==len(_jobs)
            
            for i,instance in enumerate(instances):
                assert instance.get_name() == _instances[i].get_name()
                # more general to not compare the instances 
                # this means we need to detect another way if the instance failed before 
                # and reload the new instance in the state ?
                # >> this is now done with instances_states in provider + jobs 'UNKNOWN' states detection ....
                #assert instance.get_id()   == _instances[i].get_id()
                assert instance.get_cpus() == _instances[i].get_cpus()
            for i,env in enumerate(environments):
                assert env.get_name_with_hash() == _environments[i].get_name_with_hash()
            for i,job in enumerate(jobs):
                assert job.get_hash() == _jobs[i].get_hash() # this ensures input,uploads and script are the same....
                assert job.get_config('run_script') == _jobs[i].get_config('run_script') # hash doesnt capture the args 
                assert job.get_config('run_command') == _jobs[i].get_config('run_command') # no need (dont with hash), but for symetry with run_script....
                assert job.get_config('cpus_req') == _jobs[i].get_config('cpus_req')
                assert job.get_rank() == _jobs[i].get_rank()
            return True
        except Exception as e:
            self._provider.debug(1,e)
            return False

    def transfer(self):
        return self._loaded['state'] , self._loaded['instances'] , self._loaded['environments'] , self._loaded['jobs'] , self._loaded['run_sessions'] , self._loaded['current_session']